import unittest
from analogai.deepthink.tests.functional.storage_cleaner import cleanup_storages

class BaseFunctionalTest(unittest.TestCase):
    
    def setUp(self):
        cleanup_storages()
    
    def tearDown(self):
        cleanup_storages()






from __future__ import annotations

from concurrent.futures import ThreadPoolExecutor
from enum import Enum
from typing import Iterable, Optional, Protocol


class FactExecutionMode(str, Enum):
    SEQUENTIAL = "sequential"
    PARALLEL = "parallel"


class FactTaskHandler(Protocol):
    def handle(self, segment: str) -> Optional[str]:
        ...


class FactExecutionStrategy:
    def execute(
        self,
        segments: Iterable[str],
        handler: FactTaskHandler,
    ) -> list[Optional[str]]:
        return [handler.handle(segment) for segment in segments]


class SequentialFactExecutor(FactExecutionStrategy):
    def execute(
        self,
        segments: Iterable[str],
        handler: FactTaskHandler,
    ) -> list[Optional[str]]:
        return [handler.handle(segment) for segment in segments]


class ParallelFactExecutor(FactExecutionStrategy):
    def __init__(self, max_workers: int | None = None) -> None:
        self._max_workers = max_workers

    def execute(
        self,
        segments: Iterable[str],
        handler: FactTaskHandler,
    ) -> list[Optional[str]]:
        with ThreadPoolExecutor(max_workers=self._max_workers) as executor:
            futures = [executor.submit(handler.handle, segment) for segment in segments]
            return [future.result() for future in futures]



from analogai.deepthink.application.usecases.learning.make_direct_statement.models import MakeDirectStatementRequest, MakeDirectStatementResponse
from analogai.deepthink.application.usecases.learning.make_direct_statement.ports import LearnInputPort, MakeStatementOutputPort
from analogai.deepthink.application.usecases.learning.make_direct_statement.states.context.learner_context_handler import LearnerContextHandler
from analogai.deepthink.application.usecases.learning.make_direct_statement.states.common.response_builder import LearnerResponseBuilder
from analogai.deepthink.application.usecases.learning.make_direct_statement.states.find_statement_state import FindStatementState
from analogai.deepthink.application.usecases.learning.make_direct_statement.states.create_statement_state import CreateStatementState
from analogai.deepthink.application.usecases.learning.make_direct_statement.states.update_statement_state import UpdateStatementState
from analogai.deepthink.application.exceptions.authority_exceptions import InsufficientAuthorityError
from analogai.foundation.modules.notion.notion_service import NotionService
from analogai.foundation.common.design_patterns.state_machine import StateMachine
from analogai.foundation.common.logging.app_logger import app_logger


class MakeDirectStatementUseCase(LearnInputPort, StateMachine):
    def __init__(
        self,
        statement_service,
        notion_service: NotionService,
        output_port: MakeStatementOutputPort,
        context_handler: LearnerContextHandler,
        response_builder: LearnerResponseBuilder,
    ):
        StateMachine.__init__(self, context_handler.context)
        self._statement_service = statement_service
        self._notion_service = notion_service
        self._output_port = output_port
        self._context_handler = context_handler
        self._response_builder = response_builder

        self.find_statement_state = FindStatementState(self, statement_service)
        self.update_statement_state = UpdateStatementState(self, statement_service, output_port, response_builder)
        self.create_statement_state = CreateStatementState(
            self,
            statement_service,
            notion_service,
            output_port,
            response_builder,
        )

    def execute(self, request: MakeDirectStatementRequest, output_port_override=None):
        app_logger.info(f"Learning statement: {repr(request)}")

        context_handler, response_builder = self._create_execution_context(request)
        
        previous_output_ports = self._configure_output_ports(output_port_override, response_builder)

        try:
            result = self._process_learning_request(request, context_handler, response_builder)
            app_logger.info(f"Successfully processed learning statement: {repr(request)}")
            return result
        except InsufficientAuthorityError:
            return self._handle_insufficient_authority(request, response_builder)
        finally:
            self._restore_output_ports(previous_output_ports)

    def _create_execution_context(self, request: MakeDirectStatementRequest):
        from analogai.deepthink.application.usecases.learning.make_direct_statement.states.context.learner_context import LearnerContext
        from analogai.deepthink.application.usecases.learning.make_direct_statement.states.context.learner_context_handler import LearnerContextHandler
        from analogai.deepthink.application.usecases.learning.make_direct_statement.states.common.response_builder import LearnerResponseBuilder

        context = LearnerContext()
        context_handler = LearnerContextHandler(context, self._notion_service)
        response_builder = LearnerResponseBuilder(context)
        
        return context_handler, response_builder
    
    def _configure_output_ports(self, output_port_override, response_builder):
        port = output_port_override if output_port_override else self._output_port
        
        previous_ports = {
            'update': self.update_statement_state._output_port,
            'create': self.create_statement_state._output_port
        }
        
        self.update_statement_state._output_port = port
        self.create_statement_state._output_port = port
        self.update_statement_state._response_builder = response_builder
        self.create_statement_state._response_builder = response_builder
        
        return previous_ports
    
    def _restore_output_ports(self, previous_ports):
        self.update_statement_state._output_port = previous_ports['update']
        self.create_statement_state._output_port = previous_ports['create']
    
    def _process_learning_request(self, request, context_handler, response_builder):
        context_handler.create_from_request(request)
        self.set_context(context_handler.context)
        return self._execute_state_machine()
    
    def _handle_insufficient_authority(self, request, response_builder):
        try:
            response = response_builder.build_response()
            return self._output_port.output_forbidden(response)
        except Exception:
            fallback_response = self._create_fallback_response(request)
            return self._output_port.output_forbidden(fallback_response)
    
    def _create_fallback_response(self, request):
        from analogai.foundation.common.dtos.belief_expression import BeliefExpression
        
        belief_expression = BeliefExpression(
            subject_notion_expression=request.subject_notion_expression,
            complement_notion_expression=request.complement_notion_expression,
            relation=request.relation,
            probability=request.probability,
            validity=1.0,
            modifier=request.modifier
        )
        
        return MakeDirectStatementResponse(
            user_agent=request.user_agent,
            belief_expression=belief_expression
        )

    def _execute_state_machine(self):
        if not self.context:
            raise ValueError("context should be set")
        return self.state_to_find_statement()

    def state_to_find_statement(self):
        return self.state_to(self.find_statement_state)

    def state_to_update_statement(self):
        return self.state_to(self.update_statement_state)

    def state_to_create_statement(self):
        return self.state_to(self.create_statement_state)



from .notion_enrichment_repository import NotionEnrichmentSleepRepository
from .notion_enrichment_sleeper import NotionEnrichmentSleeper

__all__ = [
    "NotionEnrichmentSleepRepository",
    "NotionEnrichmentSleeper",
]



"""Backward-compatible facade for building DeepthinkEngine via bootstrap."""

from typing import Optional

from analogai.deepthink.application.value_objects.llm_config import LlmConfig
from analogai.deepthink.application.value_objects.embedding_config import EmbeddingConfig
from analogai.deepthink.infrastructure.gateways.llm.gateway import LlmGateway
from analogai.deepthink.infrastructure.gateways.llm.enums.llm_provider import LlmProvider
from analogai.deepthink.application.value_objects.user_agent import UserAgent
from analogai.deepthink.presentation.preprocessing.previous_messages_stack import PreviousMessagesStack
from analogai.deepthink.messaging.question_answering.enums.synthesized_text_type import (
	SynthesizedTextType,
)
from analogai.deepthink.startup.bootstrap import create_deepthink_engine


class Deepthinker:
	"""Facade class to preserve the public API surface."""

	def __init__(
		self,
		llm_gateway_instance: Optional[LlmGateway] = None,
		llm_provider: Optional[LlmProvider] = None,
		llm_model: Optional[str] = None,
		llm_config: Optional[LlmConfig] = None,
		embedding_config: Optional[EmbeddingConfig] = None,
	):
		self._engine = create_deepthink_engine(
			llm_gateway_instance=llm_gateway_instance,
			llm_provider=llm_provider,
			llm_model=llm_model,
			llm_config=llm_config,
			embedding_config=embedding_config,
		)

	def close(self) -> None:
		self._engine.close()

	def deepthink(
		self,
		input_text: str,
		user_agent: Optional[UserAgent] = None,
		previous_messages_stack: Optional[PreviousMessagesStack] = None,
	) -> Optional[str]:
		return self._engine.deepthink(
			input_text=input_text,
			user_agent=user_agent,
			previous_messages_stack=previous_messages_stack,
		)

	def chat(
		self,
		input_text: str,
		user_agent: Optional[UserAgent] = None,
		previous_messages_stack: Optional[PreviousMessagesStack] = None,
	) -> Optional[str]:
		return self._engine.chat(
			input_text=input_text,
			user_agent=user_agent,
			previous_messages_stack=previous_messages_stack,
		)

	def remember(
		self,
		input_text: str,
		user_agent: Optional[UserAgent] = None,
		previous_messages_stack: Optional[PreviousMessagesStack] = None,
	) -> Optional[str]:
		return self._engine.remember(
			input_text=input_text,
			user_agent=user_agent,
			previous_messages_stack=previous_messages_stack,
		)

	def search(
		self,
		input_text: str,
		user_agent: Optional[UserAgent] = None,
		previous_messages_stack: Optional[PreviousMessagesStack] = None,
		synthesized_text_type: Optional[SynthesizedTextType] = None,
	) -> Optional[str]:
		return self._engine.search(
			input_text=input_text,
			user_agent=user_agent,
			previous_messages_stack=previous_messages_stack,
			synthesized_text_type=synthesized_text_type,
		)

	def generate_background_info(
		self,
		input_text: str,
		user_agent: Optional[UserAgent] = None,
		previous_messages_stack: Optional[PreviousMessagesStack] = None,
	) -> str:
		return self._engine.generate_background_info(
			input_text=input_text,
			user_agent=user_agent,
			previous_messages_stack=previous_messages_stack,
		)

	def get_background_info(
		self,
		input_text: str,
		user_agent: Optional[UserAgent] = None,
		previous_messages_stack: Optional[PreviousMessagesStack] = None,
	) -> str:
		return self._engine.get_background_info(
			input_text=input_text,
			user_agent=user_agent,
			previous_messages_stack=previous_messages_stack,
		)


DeepThinker = Deepthinker
Thinker = Deepthinker






from __future__ import annotations

from dataclasses import dataclass
from typing import Optional

from analogai.deepthink.infrastructure.gateways.llm.enums.llm_provider import LlmProvider
from analogai.deepthink.infrastructure.gateways.llm.gateway import LlmGateway


@dataclass(frozen=True)
class LlmConfig:
    provider: Optional[LlmProvider] = None
    model: Optional[str] = None
    gateway_instance: Optional[LlmGateway] = None



from __future__ import annotations

from typing import Iterable

from analogai.deepthink.application.value_objects.user_agent import UserAgent
from analogai.deepthink.messaging.question_answering.adapters.fact_search_adapter import (
    FactSearchAdapter,
)
from analogai.deepthink.messaging.fact.enums.fact_search_mode import FactSearchMode
from analogai.deepthink.messaging.question_answering.services.assembly.fact_retrieval_command import (
    FactRetrievalCommand,
)


class FactBackgroundInfoBuilder:
    def __init__(self, fact_retriever: FactSearchAdapter | None = None) -> None:
        self._fact_retriever = fact_retriever
        self._retrieval_command = FactRetrievalCommand(self._fact_retriever)

    def build_background(
        self,
        user_agent: UserAgent,
        question_text: str,
        entities: Iterable[str] | None,
        notions: Iterable | None,
        allow_question_facts: bool,
        search_mode: FactSearchMode,
        fact_lookup: dict[str, str] | None = None,
    ) -> dict[str, str]:
        return self._retrieval_command.execute(
            user_agent,
            question_text,
            entities,
            notions,
            allow_question_facts,
            search_mode,
            lookup=fact_lookup,
        )



import unittest
from analogai.deepthink.presentation.parsers.parsers_factory import build_message_parser


class TestParserCombinations(unittest.TestCase):
    def setUp(self):
        self.sentence_parser = build_message_parser()

    def test_sentence_combinations(self):
        combinations = [
            (
                '{"intent":"making_statement","subject":"smoke","relationship":"causes","object":"cancer","probability":1.0}',
                {"relation_name": "causes"},
            ),
            (
                '{"intent":"making_statement","subject":"socrates","relationship":"is_a","object":"human","probability":1.0}',
                {"relation_name": "is"},
            ),
            (
                '{"intent":"making_statement","location":"Tbilisi","subject":"it","relationship":"do","object":"rain","probability":1.0}',
                {"relation_name": "do", "location": "Tbilisi"},
            ),
            (
                '{"intent":"making_statement","time":"today","probability":0.5,"subject":"it","relationship":"do","object":"rain"}',
                {"from_time": "today", "probability": 0.5, "relation_name": "do"},
            ),
            (
                '{"intent":"making_statement","location":"Tbilisi","time":"tomorrow","deontic":"permitted","subject":"agent","relationship":"sell","object":"macbook pro","probability":1.0}',
                {"location": "Tbilisi", "from_time": "tomorrow", "deontic_modality": "permitted", "relation_name": "sell"},
            ),
            (
                '{"intent":"making_statement","deontic":"obligatory","subject":"you","relationship":"read","object":"books","probability":1.0}',
                {"deontic_modality": "obligatory", "relation_name": "read"},
            ),
            (
                '{"intent":"making_statement","time":"tomorrow","time_from":"14:00","time_to":"16:00","subject":"it","relationship":"do","object":"rain","probability":1.0}',
                {"from_time": "tomorrow 14:00", "to_time": "tomorrow 16:00", "relation_name": "do"},
            ),
        ]
        for text, expectations in combinations:
            with self.subTest(text=text[:60]):
                parsed_list = self.sentence_parser.parse(text)
                self.assertTrue(parsed_list, f"Expected parse for: {text[:60]}")
                parsed = parsed_list[0]
                for key, expected in expectations.items():
                    if key == "relation_name":
                        self.assertEqual(parsed.data.relation.value if parsed.data.relation else None, expected, f"key={key}")
                    else:
                        self.assertEqual(getattr(parsed.data, key), expected, f"key={key}")

    def test_sentence_returns_none_for_plain_text(self):
        for text in ["human be mortal", "smoking cause cancer", "smoke causes cancer"]:
            with self.subTest(text=text):
                self.assertEqual([], self.sentence_parser.parse(text))

    def test_sentence_without_modal_has_none_context(self):
        parsed_list = self.sentence_parser.parse('{"intent":"making_statement","subject":"socrates","relationship":"is_a","object":"human","probability":1.0}')
        self.assertTrue(parsed_list)
        parsed = parsed_list[0]
        self.assertIsNone(parsed.data.location)
        self.assertIsNone(parsed.data.deontic_modality)
        self.assertIsNone(parsed.data.from_time)
        self.assertIsNone(parsed.data.to_time)


if __name__ == "__main__":
    unittest.main()



import unittest
from datetime import datetime, timedelta
from types import SimpleNamespace
from unittest.mock import Mock

from analogai.foundation.common.enums import RelationshipType
from analogai.foundation.core.value_objects import Relation
from analogai.foundation.core.value_objects.modifier import Modifier
from analogai.foundation.core.value_objects.time import Time
from analogai.foundation.common.design_patterns.state_machine import State
from analogai.deepthink.application.usecases.learning.make_direct_statement.states.find_statement_state import FindStatementState
from analogai.deepthink.application.usecases.learning.make_direct_statement.usecase import MakeDirectStatementUseCase

class TestFindStatementState(unittest.TestCase):
    def setUp(self):
        self.state_machine = Mock(spec=MakeDirectStatementUseCase)
        self.statement_service = Mock()
        self.context = Mock(spec=['subject_notion', 'complement_notion', 'relation', 'user_agent', 'modifier', 'statement'])
        self.state_machine.context = self.context
        self.find_statement_state = FindStatementState(self.state_machine, self.statement_service)

    def test_init(self):
        self.assertEqual(self.find_statement_state.state_machine, self.state_machine)
        self.assertEqual(self.find_statement_state._statement_service, self.statement_service)
        self.assertIsInstance(self.find_statement_state, State)

    def test_execute_statement_found(self):
        statement = Mock()
        self.find_statement_state._find_statement = Mock(return_value=statement)
        self.find_statement_state._transition = Mock(return_value="next_state")
        
        result = self.find_statement_state.execute()
        
        self.find_statement_state._find_statement.assert_called_once()
        self.find_statement_state._transition.assert_called_once_with(statement)
        self.assertEqual(result, "next_state")

    def test_execute_no_statement(self):
        self.find_statement_state._find_statement = Mock(return_value=None)
        self.find_statement_state._transition = Mock(return_value="next_state")
        
        result = self.find_statement_state.execute()
        
        self.find_statement_state._find_statement.assert_called_once()
        self.find_statement_state._transition.assert_called_once_with(None)
        self.assertEqual(result, "next_state")

    def test_find_statement_success(self):
        statement = Mock()
        statement.modifier = None
        self.context.subject_notion = "source"
        self.context.complement_notion = "target"
        self.context.relation = Relation.from_type(RelationshipType.IS_A)
        self.context.user_agent = Mock()
        self.context.user_agent.user_id = "user"
        self.context.modifier = None
        self.statement_service.find.return_value = statement
        
        result = self.find_statement_state._find_statement()
        
        self.statement_service.find.assert_called_once_with(
            subject_notion="source",
            complement_notion="target",
            relation=self.context.relation,
            modifier=None
        )
        self.assertEqual(self.context.statement, statement)
        self.assertEqual(result, statement)

    def test_find_statement_none(self):
        test_relation = Relation.from_type(RelationshipType.IS_A)
        context = SimpleNamespace(
            subject_notion="source",
            complement_notion="target",
            relation=test_relation,
            user_agent=Mock(),
            modifier=None,
        )
        context.user_agent.user_id = "user"
        self.state_machine.context = context
        self.statement_service.find.return_value = None

        result = self.find_statement_state._find_statement()

        self.statement_service.find.assert_called_once_with(
            subject_notion="source",
            complement_notion="target",
            relation=test_relation,
            modifier=None
        )
        with self.assertRaises(AttributeError):
            _ = context.statement
        self.assertIsNone(result)

    def test_find_statement_with_mismatched_modifier_returns_none(self):
        today = datetime.now()
        tomorrow = today + timedelta(days=1)

        statement = Mock()
        statement.modifier = Modifier(from_time=Time(year=today.year, month=today.month, day=today.day))

        context = SimpleNamespace(
            subject_notion="source",
            complement_notion="target",
            relation=Relation.from_type(RelationshipType.IS_A),
            user_agent=Mock(),
            modifier=Modifier(from_time=Time(year=tomorrow.year, month=tomorrow.month, day=tomorrow.day)),
        )
        context.user_agent.user_id = "user"
        self.state_machine.context = context
        self.statement_service.find.return_value = statement

        result = self.find_statement_state._find_statement()

        self.statement_service.find.assert_called_once()
        self.assertIsNone(result)
        with self.assertRaises(AttributeError):
            _ = context.statement

    def test_find_statement_with_matching_modifier_returns_statement(self):
        today = datetime.now()
        modifier = Modifier(from_time=Time(year=today.year, month=today.month, day=today.day))
        
        statement = Mock()
        statement.modifier = modifier
        
        self.context.subject_notion = "source"
        self.context.complement_notion = "target"
        self.context.relation = Relation.from_type(RelationshipType.IS_A)
        self.context.user_agent = Mock()
        self.context.user_agent.user_id = "user"
        self.context.modifier = modifier
        self.statement_service.find.return_value = statement
        
        result = self.find_statement_state._find_statement()
        
        self.statement_service.find.assert_called_once()
        self.assertEqual(self.context.statement, statement)
        self.assertEqual(result, statement)

    def test_transition_no_statement(self):
        self.state_machine.state_to_create_statement.return_value = "create_state"
        
        result = self.find_statement_state._transition(None)
        
        self.state_machine.state_to_create_statement.assert_called_once()
        self.assertEqual(result, "create_state")

    def test_transition_statement_found(self):
        statement = Mock()
        self.state_machine.state_to_update_statement.return_value = "update_state"

        result = self.find_statement_state._transition(statement)

        self.state_machine.state_to_update_statement.assert_called_once()
        self.assertEqual(result, "update_state")

    def test_transition_to_create(self):
        self.state_machine.state_to_create_statement.return_value = "create_state"

        result = self.find_statement_state._transition_to_create()

        self.state_machine.state_to_create_statement.assert_called_once()
        self.assertEqual(result, "create_state")

    def test_transition_to_check_update_applicable(self):
        self.state_machine.state_to_update_statement.return_value = "update_state"

        result = self.find_statement_state._transition_to_check_update_applicable()

        self.state_machine.state_to_update_statement.assert_called_once()
        self.assertEqual(result, "update_state")

if __name__ == '__main__':
    unittest.main()






from typing import Optional, Callable

from analogai.deepthink.application.value_objects.user_agent import UserAgent
from analogai.deepthink.presentation.preprocessing.previous_messages_stack import PreviousMessagesStack
from analogai.deepthink.internal.processors.deepthinker_processor import DeepthinkPipelineProcessor
from analogai.deepthink.messaging.question_answering.enums.synthesized_text_type import (
    SynthesizedTextType,
)


class DeepthinkEngine:
    def __init__(
        self,
        processor: DeepthinkPipelineProcessor,
        on_close: Optional[Callable[[], None]] = None,
    ):
        self._processor = processor
        self._on_close = on_close

    def close(self) -> None:
        if self._on_close is not None:
            self._on_close()

    def deepthink(
        self,
        input_text: str,
        user_agent: Optional[UserAgent] = None,
        previous_messages_stack: Optional[PreviousMessagesStack] = None,
    ) -> Optional[str]:
        return self._processor.deepthink(
            input_text=input_text,
            user_agent=user_agent,
            previous_messages_stack=previous_messages_stack,
        )

    def chat(
        self,
        input_text: str,
        user_agent: Optional[UserAgent] = None,
        previous_messages_stack: Optional[PreviousMessagesStack] = None,
    ) -> Optional[str]:
        return self._processor.chat(
            input_text=input_text,
            user_agent=user_agent,
            previous_messages_stack=previous_messages_stack,
        )

    def remember(
        self,
        input_text: str,
        user_agent: Optional[UserAgent] = None,
        previous_messages_stack: Optional[PreviousMessagesStack] = None,
    ) -> Optional[str]:
        return self._processor.remember(
            input_text=input_text,
            user_agent=user_agent,
            previous_messages_stack=previous_messages_stack,
        )

    def search(
        self,
        input_text: str,
        user_agent: Optional[UserAgent] = None,
        previous_messages_stack: Optional[PreviousMessagesStack] = None,
        synthesized_text_type: Optional[SynthesizedTextType] = None,
    ) -> Optional[str]:
        return self._processor.search(
            input_text=input_text,
            user_agent=user_agent,
            previous_messages_stack=previous_messages_stack,
            synthesized_text_type=synthesized_text_type,
        )

    def generate_background_info(
        self,
        input_text: str,
        user_agent: Optional[UserAgent] = None,
        previous_messages_stack: Optional[PreviousMessagesStack] = None,
    ) -> str:
        return self._processor.generate_background_info(
            input_text=input_text,
            user_agent=user_agent,
            previous_messages_stack=previous_messages_stack,
        )

    def get_background_info(
        self,
        input_text: str,
        user_agent: Optional[UserAgent] = None,
        previous_messages_stack: Optional[PreviousMessagesStack] = None,
    ) -> str:
        return self._processor.get_background_info(
            input_text=input_text,
            user_agent=user_agent,
            previous_messages_stack=previous_messages_stack,
        )


DeepThinker = DeepthinkEngine
Thinker = DeepthinkEngine



from analogai.deepthink.application.usecases.learning.make_conditional_statement.usecase import MakeConditionalStatementUseCase
from analogai.deepthink.application.usecases.learning.make_conditional_statement.models import MakeConditionalStatementRequest, MakeConditionalStatementResponse
from analogai.deepthink.application.usecases.learning.make_conditional_statement.ports import MakeConditionalStatementOutputPort

__all__ = [
    "MakeConditionalStatementUseCase",
    "MakeConditionalStatementRequest",
    "MakeConditionalStatementResponse",
    "MakeConditionalStatementOutputPort",
]



from typing import Optional

from analogai.foundation.core.value_objects.modifier import Modifier
from analogai.foundation.core.value_objects import Relation
from analogai.foundation.common.dtos.notion_expression import NotionExpression
from analogai.deepthink.application.value_objects.user_agent import UserAgent


class LearnerContext:
    def __init__(self):
        self.user_agent: UserAgent | None = None
        self.subject_notion = None
        self.complement_notion = None
        self.subject_notion_expression: NotionExpression | None = None
        self.complement_notion_expression: NotionExpression | None = None
        self.relation: Optional[Relation] = None
        self.proposed_probability = None
        self.proposed_validity = None
        self.proposed_quantity: Optional[float] = None
        self.statement = None
        self.modifier: Optional[Modifier] = None
        self.statement_name: Optional[str] = None
        self.complement_caption_rest: Optional[str] = None



from .utils.fact_context import get_current_turn_id, set_current_turn_id, use_turn_id
from .services.fact_context_assembler import FactContextBuilder
from .factories.fact_domain_service_factory import FactDomainServiceFactory
from .services.fact_context_service import FactContextWindowService
from .factories.fact_factory import FactServiceFactory
from .fact import Fact
from .services.fact_response_decorator import FactResponseDecorator
from .repositories.fact_repository import FactRepositoryPort
from .repositories.infrastructure.fact_repository import (
    InMemoryFactRepository,
    MongoFactRepository,
)
from .services.fact_service import (
    FactServicePort,
    InMemoryFactService,
    MongoFactService,
    RepositoryBackedFactService,
)
from .services.fact_statement_service import FactStatementService
from .services.fact_hypothetical_statement_service import HypotheticalStatementLinkingService
from .enums.fact_reference_type import FactReferenceType
from .services.fact_turn_recorder import FactTurnTracker

__all__ = [
    "get_current_turn_id",
    "set_current_turn_id",
    "use_turn_id",
    "FactContextBuilder",
    "FactDomainServiceFactory",
    "FactContextWindowService",
    "FactServiceFactory",
    "Fact",
    "FactResponseDecorator",
    "FactRepositoryPort",
    "InMemoryFactRepository",
    "MongoFactRepository",
    "FactServicePort",
    "InMemoryFactService",
    "MongoFactService",
    "RepositoryBackedFactService",
    "FactStatementService",
    "HypotheticalStatementLinkingService",
    "FactReferenceType",
    "FactTurnTracker",
]


from typing import List
from analogai.foundation.common.logging.app_logger import app_logger
from .postprocessors.pronoun_postprocessior import postprocesses_pronouns
from .postprocessors.general_postprocessor import capitalize_first

class BasePostprocessorHandler:
    def process(self, text: str) -> str:
        raise NotImplementedError

class PronounPostprocessorHandler(BasePostprocessorHandler):
    def process(self, text: str) -> str:
        return postprocesses_pronouns(text)

class GeneralPostprocessorHandler(BasePostprocessorHandler):
    def process(self, text: str) -> str:
        return capitalize_first(text)

def create_postprocessor_chain() -> List[BasePostprocessorHandler]:
    postprocessors: List[BasePostprocessorHandler] = [
        PronounPostprocessorHandler(),
        GeneralPostprocessorHandler()
    ]
    return postprocessors



from __future__ import annotations

from analogai.deepthink.messaging.fact.services.fact_context_assembler import FactContextBuilder


class FactResponseDecorator:
    def __init__(self, assembler: FactContextBuilder):
        self._assembler = assembler

    def decorate_belief_line(self, belief_line: str, belief) -> str:
        extra = self._assembler.build_belief_context(belief)
        if not extra:
            return belief_line
        return f"{belief_line} ({extra})"



from __future__ import annotations

from pathlib import Path
from typing import List

from analogai.deepthink.messaging.question_answering.enums.synthesized_text_type import (
    SynthesizedTextType,
)

class QASynthesizerPromptComposer:
    def build_system_prompt(self) -> str:
        return ""

    def build_user_message(
        self,
        question: str,
        entities: List[str],
        belief_block: str,
    ) -> str:
        entities_text = ", ".join(entities) if entities else "(none detected)"
        belief_text = belief_block or "(no matching beliefs found)"
        return (
            f"Question: {question}\n"
            f"Entities: {entities_text}\n"
            "Belief context:\n"
            f"{belief_text}\n\n"
            "Answer:"
        )

    def build_synthesis_prompt(
        self,
        question: str,
        tasks: str,
        facts: str,
        qa_history: str,
        synthesized_text_type: SynthesizedTextType,
    ) -> str:
        template = _load_synthesizer_prompt_template(synthesized_text_type)
        return template.format(
            question=question,
            tasks=tasks,
            facts=facts,
            qa_history=qa_history,
        )


def _load_synthesizer_prompt_template(synthesized_text_type: SynthesizedTextType) -> str:
    prompts_dir = Path(__file__).resolve().parents[2] / "infrastructure" / "prompts"
    filename_map = {
        SynthesizedTextType.EXACT_MATCH: "qa_synthesizer_prompt_exact_match.txt",
        SynthesizedTextType.DETAILED: "qa_synthesizer_prompt_detailed.txt",
    }
    prompt_path = prompts_dir / filename_map[synthesized_text_type]
    if not prompt_path.exists():
        raise FileNotFoundError(f"Prompt file not found: {prompt_path}")
    return prompt_path.read_text(encoding="utf-8")



from __future__ import annotations

from dataclasses import dataclass
from typing import Optional
from enum import Enum

from analogai.foundation.infrastructure.services.embedding_service import EmbeddingService
from analogai.foundation.common.enums.embedding_type import EmbeddingType  # type: ignore[import-not-found]


@dataclass(frozen=True)
class EmbeddingConfig:
    provider_instance: Optional[EmbeddingService] = None
    embed_type: Optional[EmbeddingType] = None
    provider: Optional[str | Enum] = None
    model: Optional[str] = None

    def to_overrides(self) -> dict[str, str]:
        overrides: dict[str, str] = {}
        if self.embed_type:
            overrides["embed.type"] = self.embed_type.value
        if self.provider:
            provider_value = (
                self.provider.value
                if isinstance(self.provider, Enum)
                else self.provider
            )
            overrides["embed.provider"] = provider_value
        if self.model:
            overrides["embed.model"] = self.model
        return overrides



from __future__ import annotations

from typing import Iterable, List, Tuple

from analogai.deepthink.application.value_objects.user_agent import UserAgent
from analogai.deepthink.messaging.question_answering.services.formatting.fact_formatter import (
    QuestionFactFormatter,
)
from analogai.deepthink.messaging.question_answering.adapters.belief_search_adapter import (
    BeliefSearchAdapter,
)
from analogai.deepthink.messaging.question_answering.adapters.notion_search_adapter import (
    NotionSearchAdapter,
)


class BeliefBackgroundInfoBuilder:
    def __init__(
        self,
        belief_retriever: BeliefSearchAdapter,
        notion_retriever: NotionSearchAdapter,
        belief_formatter: QuestionFactFormatter,
    ) -> None:
        self._belief_retriever = belief_retriever
        self._notion_retriever = notion_retriever
        self._belief_formatter = belief_formatter

    def build_background(
        self,
        user_agent: UserAgent,
        entities: Iterable[str],
    ) -> Tuple[List, List, List[str], str, int]:
        if not entities:
            return [], [], [], "", 0
        beliefs = self._belief_retriever.search_beliefs(user_agent, entities)
        notions = self._notion_retriever.search_notions(user_agent, entities)
        notions = self._notion_retriever.expand_notions_from_beliefs(user_agent, notions, beliefs)
        notion_text = self._belief_formatter.format_notions(notions)
        notion_lines = [line for line in notion_text.splitlines() if line.strip()] if notion_text else []
        belief_block = self._belief_formatter.build_fact_block_text(beliefs)
        belief_count = self._count_serialized_beliefs(beliefs)
        return beliefs, notions, notion_lines, belief_block, belief_count

    def _count_serialized_beliefs(self, beliefs: Iterable) -> int:
        count = 0
        for belief in beliefs or []:
            if self._belief_formatter.format_fact(belief):
                count += 1
        return count



from abc import ABC, abstractmethod
from analogai.deepthink.application.usecases.learning.learn_statement.models import LearnStatementRequest


class LearnStatementInputPort(ABC):
    @abstractmethod
    def execute(self, request: LearnStatementRequest):
        pass


class LearnStatementOutputPort(ABC):
    pass



import json
import unittest

from analogai.deepthink.application.value_objects.user_agent import UserAgent
from analogai.deepthink.infrastructure.gateways.llm.gateway import LlmGateway
from analogai.deepthink.infrastructure.gateways.llm.enums.llm_provider import LlmProvider
from analogai.deepthink.infrastructure.gateways.llm.llm_factory import LlmGatewayFactory
from analogai.deepthink.infrastructure.gateways.llm.llm_gateway_provider_factory import (
    LlmGatewayProviderFactory,
)
from analogai.deepthink.presentation.preprocessing.previous_messages_stack import PreviousMessagesStack
from analogai.deepthink.tests.functional.storage_cleaner import cleanup_storages
from analogai.deepthink.startup.bootstrap import create_deepthink_engine


class TestThinkerPreviousMessageStackIntegration(unittest.TestCase):
    def setUp(self):
        cleanup_storages()
        class _TestGateway(LlmGateway):
            def __init__(self):
                self._prompt = ""
                self._model = None

            def set_system_prompt(self, prompt: str) -> None:
                self._prompt = prompt

            def set_model(self, model_id: str) -> None:
                self._model = model_id

            def generate_completion(self, user_message: str) -> str:
                sentence_text = user_message.strip()
                last_line = sentence_text.splitlines()[-1] if sentence_text else ""
                if last_line.startswith("Sentence:"):
                    sentence_text = last_line.replace("Sentence:", "", 1).strip()
                history_subject = _extract_last_subject(user_message)
                if history_subject:
                    sentence_text = _resolve_pronouns(sentence_text, history_subject)
                parsed = {
                    "intent_type": "making_statement",
                    "subject": sentence_text,
                    "predicate": {"relation": "is", "complement": "unknown"},
                }
                return json.dumps({"sentences": [parsed]})

            async def generate_streaming_completion(self, user_message: str):
                yield self.generate_completion(user_message)

        self._prev_provider_create_gateway = LlmGatewayProviderFactory.create_gateway

        def _create_gateway_for_test(
            cls,
            gateway_type=None,
            system_prompt: str | None = None,
            max_tokens=None,
            model: str | None = None,
        ):
            gateway = _TestGateway()
            if model:
                gateway.set_model(model)
            if system_prompt is not None:
                gateway.set_system_prompt(system_prompt)
            return gateway

        LlmGatewayProviderFactory.create_gateway = classmethod(_create_gateway_for_test)

        self._prev_gateway_type = LlmGatewayFactory._default_gateway_type
        self._prev_gateway_instance = LlmGatewayFactory._default_gateway_instance
        LlmGatewayFactory.set_default_gateway_type(LlmProvider.OPENAI)
        LlmGatewayFactory.set_default_gateway_instance(_TestGateway())

        self._deepthink = create_deepthink_engine(
            llm_gateway_instance=_TestGateway(),
            llm_provider=LlmProvider.OPENAI,
            llm_model="test-model",
        )

    def tearDown(self):
        LlmGatewayProviderFactory.create_gateway = self._prev_provider_create_gateway
        LlmGatewayFactory.set_default_gateway_type(self._prev_gateway_type)
        LlmGatewayFactory.set_default_gateway_instance(self._prev_gateway_instance)
        cleanup_storages()

    def test_conversation_saved_per_user_after_think(self):
        user_id = "user-a"
        agent_id = "agent-1"
        stack = PreviousMessagesStack()
        user_agent = UserAgent(user_id=user_id, agent_id=agent_id, user_authority=1.0)
        self._deepthink.deepthink("socrates is human", user_agent=user_agent, previous_messages_stack=stack)
        history = list(stack.get_history())
        self.assertEqual(len(history), 2)
        self.assertIn("socrates", history[0].lower())
        self.assertIn("human", history[0].lower())

    def test_conversation_adapted_to_stack_and_used_for_pronoun_resolution(self):
        user_id = "user-b"
        agent_id = "agent-1"
        stack = PreviousMessagesStack()
        user_agent = UserAgent(user_id=user_id, agent_id=agent_id, user_authority=1.0)
        self._deepthink.deepthink("john is mortal", user_agent=user_agent, previous_messages_stack=stack)
        result = self._deepthink.deepthink("he is philosopher", user_agent=user_agent, previous_messages_stack=stack)
        self.assertIsNotNone(result)
        self.assertIn("john", result.lower())
        self.assertNotIn(" he ", f" {result.lower()} ")
        history = list(stack.get_history())
        self.assertEqual(len(history), 4)

    def test_conversations_isolated_per_user(self):
        stack_user1 = PreviousMessagesStack()
        stack_user2 = PreviousMessagesStack()
        user_agent_1 = UserAgent(user_id="user-1", agent_id="agent-1", user_authority=1.0)
        user_agent_2 = UserAgent(user_id="user-2", agent_id="agent-1", user_authority=1.0)
        self._deepthink.deepthink("alice likes apples", user_agent=user_agent_1, previous_messages_stack=stack_user1)
        self._deepthink.deepthink("bob likes oranges", user_agent=user_agent_2, previous_messages_stack=stack_user2)
        history_user1 = list(stack_user1.get_history())
        history_user2 = list(stack_user2.get_history())
        self.assertEqual(history_user1[0], "alice likes apples")
        self.assertEqual(history_user2[0], "bob likes oranges")

    def test_conversation_adapted_to_stack_returns_persisted_history(self):
        user_id = "user-c"
        agent_id = "agent-1"
        stack = PreviousMessagesStack()
        user_agent = UserAgent(user_id=user_id, agent_id=agent_id, user_authority=1.0)
        self._deepthink.deepthink("maria reads books", user_agent=user_agent, previous_messages_stack=stack)
        history = list(stack.get_history())
        self.assertGreater(len(history), 0)
        self.assertIn("maria reads books", history)


def _extract_last_subject(message: str) -> str | None:
    for line in reversed(message.splitlines()):
        line = line.strip()
        if line.lower().startswith("q:"):
            content = line[2:].strip()
            for sep in (" is ", " are "):
                if sep in content:
                    return content.split(sep)[0].strip()
            return content.split()[0] if content else None
    return None


def _resolve_pronouns(sentence: str, subject: str) -> str:
    replacements = {
        " he ": f" {subject} ",
        " she ": f" {subject} ",
        " they ": f" {subject} ",
        " him ": f" {subject} ",
        " her ": f" {subject} ",
    }
    padded = f" {sentence.strip()} "
    lowered = padded.lower()
    for token, replacement in replacements.items():
        if token in lowered:
            lowered = lowered.replace(token, replacement)
    return lowered.strip()


if __name__ == "__main__":
    unittest.main()



from __future__ import annotations

from typing import Iterable, List


class InterpreterRegistry:
    def __init__(self, interpreters: Iterable) -> None:
        self._interpreters: List = list(interpreters)

    def register(self, interpreter) -> None:
        self._interpreters.append(interpreter)

    def all(self) -> List:
        return list(self._interpreters)



from enum import Enum


class IntentType(Enum):
    MAKING_STATEMENT = "making_statement"
    MAKING_CAUSAL_STATEMENT = "making_conditional_statement"
    MAKING_NOTION_ATTRIBUTE_STATEMENT = "making_notion_attribute_statement"



from .chat_flow_factory import ChatFlowBuilder
from .command_factory import ChatCommandFactory
from .pipeline_factory import MessagePipelineFactory
from .user_agent_resolver_factory import UserAgentResolverFactory

__all__ = [
    "ChatFlowBuilder",
    "ChatCommandFactory",
    "MessagePipelineFactory",
    "UserAgentResolverFactory",
]



from typing import Optional, Any
from analogai.deepthink.presentation.factories.intents_factory import IntentsFactory
from analogai.deepthink.presentation.intent_type import IntentType
from analogai.deepthink.presentation.presenters.presenter_response import PresenterResponse


class IntentsProcessor:
	def __init__(self, intents_factory: IntentsFactory):
		self.intents_factory = intents_factory

	def execute(self, intent_type: Optional[IntentType], request: Any) -> Optional[PresenterResponse]:
		return self.intents_factory.get_intent(intent_type).execute(request)





from analogai.deepthink.presentation.presenters.utils.plural_utils import (
    pluralize_category,
    get_plural_relationship,
)
from analogai.deepthink.presentation.presenters.utils.list_format_utils import (
    format_items_and,
    format_items_and_capitalized,
)
from analogai.deepthink.presentation.presenters.utils.belief_format_utils import (
    format_grouped_statement,
    format_single_belief_statement,
)

__all__ = [
    "pluralize_category",
    "get_plural_relationship",
    "format_items_and",
    "format_items_and_capitalized",
    "format_grouped_statement",
    "format_single_belief_statement",
]



from dataclasses import dataclass, field
from typing import Optional, List, TYPE_CHECKING, TypedDict

if TYPE_CHECKING:
    from analogai.foundation.common.enums.certainty import Certainty


class AttributeDict(TypedDict, total=False):
    tag: str
    value: str
    unit: Optional[str]


@dataclass
class ParsedData:
    subject: Optional[str] = None
    complement: Optional[str] = None
    subject_attributes: Optional[List[AttributeDict]] = None
    complement_attributes: Optional[List[AttributeDict]] = None
    relationship_phrase: Optional[str] = None
    location: Optional[str] = None
    from_time: Optional[str] = None
    to_time: Optional[str] = None
    time: Optional[str] = None
    probability: Optional[float] = None
    certainty: Optional["Certainty"] = None
    attributes: Optional[List[AttributeDict]] = None
    quantity: Optional[str] = None
    deontic_modality: Optional[str] = None
    causes: Optional[List["ParsedData"]] = None
    effects: Optional[List["ParsedData"]] = None


@dataclass
class ParsedSentence:
    intent_type: Optional[str] = None
    data: Optional[ParsedData] = None
    
    def __repr__(self):
        return f"ParsedSentence(intent_type={repr(self.intent_type)}, data={repr(self.data)})"



from __future__ import annotations

from typing import Any, Dict, Optional

from analogai.deepthink.messaging.fact.fact import Fact


class FactStorageMapper:
    @staticmethod
    def to_document(fact: Fact) -> Dict[str, Any]:
        return {
            "_id": fact.id,
            "turn_id": fact.turn_id,
            "statement_id": fact.statement_id,
            "notion_id": fact.notion_id,
            "reference_type": fact.reference_type,
            "user_id": fact.user_id,
            "agent_id": fact.agent_id,
            "text": fact.text,
            "embedding": fact.embedding,
            "prev_id": fact.prev_id,
            "next_id": fact.next_id,
        }

    @staticmethod
    def from_document(doc: Optional[Dict[str, Any]]) -> Optional[Fact]:
        if not doc:
            return None
        embedding = doc.get("embedding")
        if embedding is not None and not isinstance(embedding, list):
            embedding = None
        return Fact(
            id=str(doc.get("_id")),
            turn_id=doc.get("turn_id"),
            statement_id=doc.get("statement_id"),
            notion_id=doc.get("notion_id"),
            reference_type=doc.get("reference_type"),
            user_id=str(doc.get("user_id")),
            agent_id=str(doc.get("agent_id")),
            text=str(doc.get("text") or ""),
            embedding=embedding,
            prev_id=doc.get("prev_id"),
            next_id=doc.get("next_id"),
        )



import re
from abc import ABC, abstractmethod
from analogai.deepthink.presentation.enums.pronoun_placeholder import PronounPlaceholder


def _lowercase_first_letter(text: str) -> str:
    if not text or not text.strip():
        return text
    stripped = text.lstrip()
    if not stripped:
        return text
    first = stripped[0].lower()
    return text[: len(text) - len(stripped)] + first + stripped[1:]


def _uppercase_first_letter(text: str) -> str:
    if not text or not text.strip():
        return text
    stripped = text.lstrip()
    if not stripped:
        return text
    first = stripped[0].upper()
    return text[: len(text) - len(stripped)] + first + stripped[1:]


class PronounReplacementStep(ABC):
    @abstractmethod
    def apply(self, text: str) -> str:
        pass


class PlaceholderPronounReplacer(PronounReplacementStep):
    def apply(self, text: str) -> str:
        company = PronounPlaceholder.COMPANY_PLACEHOLDER.value
        agent = PronounPlaceholder.AGENT_PLACEHOLDER.value
        respondent = PronounPlaceholder.RESPONDENT_PLACEHOLDER.value

        result = text
        result = re.sub(rf"{respondent}[:#-][\w-]+", respondent, result, flags=re.IGNORECASE)
        patterns = [
            (f"an {agent}", "I"),
            (f"the {agent}", "I"),
            (f"a {respondent}", "you"),
            (f"the {respondent}", "you"),
            ("the company", "company"),
            ("a company", "company"),
            (f"{agent} is", "I am"),
            (f"{agent}'s", "my"),
            (f"{agent}s", "I"),
            (agent, "I"),
            ("I does", "I do"),
            (f"{respondent} is", "you are"),
            (f"{respondent}'s", "your"),
            (f"{respondent}s", "you"),
            (respondent, "you"),
            ("you does", "you do"),
            (f"{company} is", "we are"),
            (f"{company}'s", "our"),
            (f"{company} does", "we do"),
            (company, "we"),
        ]
        for pattern, replacement in patterns:
            result = re.sub(re.escape(pattern), replacement, result, flags=re.IGNORECASE)
        return result


class AgentMessagePronounPostprocessor:
    def __init__(self, replacer: PronounReplacementStep):
        self._replacer = replacer

    def format(self, text: str) -> str:
        if not text or not text.strip():
            return text
        lower = _lowercase_first_letter(text)
        replaced = self._replacer.apply(lower)
        return _uppercase_first_letter(replaced)


_default_postprocessor = AgentMessagePronounPostprocessor(PlaceholderPronounReplacer())


def postprocesses_pronouns(text: str) -> str:
    return _default_postprocessor.format(text)



from typing import Optional
from analogai.deepthink.features.notion_enrichment.ports import NotionCreatedObserverPort

_observer: Optional[NotionCreatedObserverPort] = None


def get_notion_enrichment_observer() -> Optional[NotionCreatedObserverPort]:
    return _observer


def register_notion_enrichment_observer(observer: NotionCreatedObserverPort) -> None:
    global _observer
    _observer = observer



from enum import Enum


class LlmProvider(str, Enum):
    XAI = "xai"
    ANTHROPIC = "anthropic"
    OPENAI = "openai"
    AZURE = "azure"
    AZURE_OPENAI = "azure_openai"
    BEDROCK = "bedrock"
    FAKE = "fake"



from analogai.deepthink.presentation.interpreters.collection.conditional_statement_interpreter import ConditionalStatementInterpreter
from analogai.deepthink.presentation.interpreters.collection.statement_learning_interpreter import StatementLearningInterpreter
from analogai.deepthink.presentation.interpreters.collection.notion_attribute_statement_interpreter import (
    NotionAttributeStatementInterpreter,
)
from analogai.deepthink.presentation.interpreters.interpreter_registry import InterpreterRegistry


class InterpretersFactory:
    def create_all(self):
        return [
            ConditionalStatementInterpreter(),
            NotionAttributeStatementInterpreter(),
            StatementLearningInterpreter(),
        ]

    def create_registry(self) -> InterpreterRegistry:
        return InterpreterRegistry(self.create_all())







from analogai.deepthink.application.utils.certainty.certainty_comparison_util import check_update_statement_applicable
from analogai.deepthink.application.usecases.learning.make_direct_statement.ports import MakeStatementOutputPort
from analogai.foundation.common.design_patterns.state_machine import State
from analogai.foundation.common.logging.app_logger import app_logger
from analogai.deepthink.application.usecases.learning.make_direct_statement.utils.modifier_util import modifiers_match
from .common.response_builder import LearnerResponseBuilder


class UpdateStatementState(State):
    def __init__(self, state_machine, statement_service, output_port: MakeStatementOutputPort,
                 response_builder: LearnerResponseBuilder):
        super().__init__(state_machine)
        self._statement_service = statement_service
        self._response_builder = response_builder
        self._output_port = output_port

    def execute(self):
        if self.context.statement is None:
            app_logger.warning(
                "Skipping update check because no existing statement is present."
            )
            return self._build_and_output_not_applicable_response()

        app_logger.info(
            "Checking if statement update is applicable: "
            f"existing probability={self.context.statement.probability}, "
            f"proposed={self.context.proposed_probability}"
        )
        
        if not self._validate_modifier_matches():
            app_logger.info("Modifiers don't match, update not applicable")
            return self._build_and_output_not_applicable_response()
        
        if self._check_is_statement_already_known():
            app_logger.info("Statement already known with same probability")
            return self._build_and_output_statement_already_known()
        
        if self._check_should_update_statement():
            app_logger.info("Updating statement probability and validity")
            return self._apply_update_and_output_success()
        
        app_logger.info("Update not applicable based on certainty comparison")
        return self._build_and_output_not_applicable_response()

    def _check_should_update_statement(self):
        if self.context.statement is None:
            return False
        return check_update_statement_applicable(
            self.context.statement.probability,
            self.context.proposed_probability,
            self.context.statement.validity,
            self.context.proposed_validity
        )

    def _check_is_statement_already_known(self):
        if self.context.statement is None:
            return False
        return self.context.statement.probability == self.context.proposed_probability

    def _apply_update_and_output_success(self):
        if self.context.statement:
            updated_statement = self._statement_service.update_probability(
                self.context.statement,
                self.context.proposed_probability
            )
            updated_statement = self._statement_service.update_validity(
                updated_statement,
                self.context.proposed_validity
            )
            self.context.statement = updated_statement
        response = self._response_builder.build_response()
        return self._output_port.output_statement_update_success(response)

    def _build_and_output_statement_already_known(self):
        response = self._response_builder.build_response()
        return self._output_port.output_statement_already_known(response)

    def _build_and_output_not_applicable_response(self):
        response = self._response_builder.build_response()
        return self._output_port.output_update_not_applicable(response)

    def _validate_modifier_matches(self) -> bool:
        statement_modal = self.context.statement.modifier if self.context.statement else None
        request_modal = self.context.modifier
        return modifiers_match(statement_modal, request_modal)




from __future__ import annotations

from typing import Optional, Dict, Tuple

from analogai.deepthink.application.value_objects.user_agent import UserAgent
from analogai.deepthink.settings import config as app_config
from analogai.foundation.common.utils.env_utils import resolve_cpu_count
from analogai.foundation.common.logging.app_logger import app_logger
from analogai.foundation.extensions.conversation.services.conversation_service import ConversationService
from analogai.deepthink.messaging.commands import GenerateBackgroundInfoCommand, RememberCommand, SearchCommand
from analogai.deepthink.messaging.flows.chat_flow import ChatFlow
from analogai.deepthink.internal.processors.chat_synthesizer import ChatResponseSynthesizer
from analogai.deepthink.internal.processors.fact_execution_strategy import FactExecutionMode
from analogai.deepthink.internal.processors.fact_task_runner import FactTaskRunner
from analogai.deepthink.internal.processors.large_text_processor import SentenceBatcher
from analogai.deepthink.messaging.fact.services.fact_service import FactServicePort
from analogai.deepthink.messaging.fact.services.fact_turn_recorder import FactTurnTracker
from analogai.deepthink.messaging.text.sentence_chunker import split_into_sentences
from analogai.deepthink.messaging.identity.user_agent_resolver import UserAgentResolver
from analogai.deepthink.messaging.filters.math_content_filter import RememberEligibilityPolicy
from analogai.deepthink.presentation.preprocessing.previous_messages_stack import PreviousMessagesStack
from analogai.deepthink.messaging.question_answering.enums.synthesized_text_type import (
    SynthesizedTextType,
)


class _RememberFactHandler:
    def __init__(
        self,
        processor: "DeepthinkPipelineProcessor",
        user_agent: UserAgent,
        stack_snapshot: PreviousMessagesStack,
    ) -> None:
        self._processor = processor
        self._user_agent = user_agent
        self._stack_snapshot = stack_snapshot

    def handle(self, segment: str) -> Optional[str]:
        return self._processor._remember_fact(
            segment,
            self._user_agent,
            self._stack_snapshot,
        )


class _BackgroundInfoHandler:
    def __init__(self, command: GenerateBackgroundInfoCommand, user_agent: UserAgent) -> None:
        self._command = command
        self._user_agent = user_agent

    def handle(self, segment: str) -> Optional[str]:
        return self._command.execute(segment, self._user_agent)


class DeepthinkPipelineProcessor:
    def __init__(
        self,
        chat_flow: ChatFlow,
        remember_command: RememberCommand,
        search_command: SearchCommand,
        background_info_command: GenerateBackgroundInfoCommand,
        user_agent_resolver: UserAgentResolver,
        large_text_processor: SentenceBatcher,
        fact_service: FactServicePort,
        turn_recorder: FactTurnTracker,
        conversation_service: Optional[ConversationService] = None,
        chat_synthesizer: Optional[ChatResponseSynthesizer] = None,
        remember_policy: Optional[RememberEligibilityPolicy] = None,
    ):
        self._chat_flow = chat_flow
        self._remember_command = remember_command
        self._search_command = search_command
        self._background_info_command = background_info_command
        self._user_agent_resolver = user_agent_resolver
        self._large_text_processor = large_text_processor
        self._fact_service = fact_service
        self._turn_recorder = turn_recorder
        self._conversation_service = conversation_service
        self._chat_synthesizer = chat_synthesizer
        self._fact_task_runner = FactTaskRunner()
        self._internal_stacks: Dict[Tuple[str, str], PreviousMessagesStack] = {}
        self._remember_policy = remember_policy or RememberEligibilityPolicy()

    def deepthink(
        self,
        input_text: str,
        user_agent: Optional[UserAgent] = None,
        previous_messages_stack: Optional[PreviousMessagesStack] = None,
    ) -> Optional[str]:
        return self.chat(
            input_text=input_text,
            user_agent=user_agent,
            previous_messages_stack=previous_messages_stack,
        )

    def chat(
        self,
        input_text: str,
        user_agent: Optional[UserAgent] = None,
        previous_messages_stack: Optional[PreviousMessagesStack] = None,
    ) -> Optional[str]:
        resolved_user_agent = self._resolve_user_agent(user_agent)
        stack = self._resolve_prev_message_stack(previous_messages_stack, resolved_user_agent)
        sentences = split_into_sentences(input_text)
        if not sentences:
            return None
        statements: list[str] = []
        questions: list[str] = []
        for sentence in sentences:
            cleaned = sentence.strip()
            if not cleaned:
                continue
            if self._chat_flow.detect_question(cleaned):
                questions.append(cleaned)
            else:
                statements.append(cleaned)

        statement_text = " ".join(statements).strip() if statements else None
        question_text = " ".join(questions).strip() if questions else None

        statement_response: Optional[str] = None
        question_response: Optional[str] = None
        if statement_text:
            with self._turn_recorder.record_user_turn(statement_text, resolved_user_agent):
                statement_results = self._remember_facts(
                    statement_text,
                    resolved_user_agent,
                    stack,
                )
            statement_response = _aggregate_responses(statement_results)
            stack.add_user(statement_text)
            if statement_response:
                stack.add_agent(statement_response)
        if question_text:
            question_response = self._chat_flow.handle_question(
                question_text,
                resolved_user_agent,
                stack,
            )

        combined = " ".join(
            part for part in [statement_response, question_response] if part
        ).strip()
        final_response: Optional[str]
        if combined:
            final_response = self._apply_chat_synthesizer(combined)
        elif question_response is not None:
            final_response = self._apply_chat_synthesizer(question_response)
        elif statement_response is not None:
            final_response = self._apply_chat_synthesizer(statement_response)
        else:
            final_response = input_text.strip() or None

        self._record_conversation_turn(resolved_user_agent, input_text, final_response)
        return final_response

    def remember(
        self,
        input_text: str,
        user_agent: Optional[UserAgent] = None,
        previous_messages_stack: Optional[PreviousMessagesStack] = None,
    ) -> Optional[str]:
        resolved_user_agent = self._resolve_user_agent(user_agent)
        stack = self._resolve_prev_message_stack(previous_messages_stack, resolved_user_agent)
        with self._turn_recorder.record_user_turn(input_text, resolved_user_agent):
            results = self._remember_facts(
                input_text,
                resolved_user_agent,
                stack,
            )
        response_text = results[-1] if results else None
        stack.add_user(input_text)
        if response_text:
            stack.add_agent(response_text)
        self._record_conversation_turn(resolved_user_agent, input_text, response_text)
        return response_text

    def _remember_facts(
        self,
        input_text: str,
        user_agent: UserAgent,
        stack: PreviousMessagesStack,
    ) -> list[Optional[str]]:
        stack_snapshot = self._clone_stack(stack)
        segments = self._large_text_processor.chunk_input(input_text)
        handler = _RememberFactHandler(self, user_agent, stack_snapshot)
        multithread_enabled = app_config.get(
            "pipelines.multithread_remember_enabled",
            True,
        )
        max_workers = _resolve_remember_max_workers()
        return self._fact_task_runner.run(
            segments,
            handler,
            mode=(
                FactExecutionMode.PARALLEL
                if multithread_enabled
                else FactExecutionMode.SEQUENTIAL
            ),
            max_workers=max_workers,
        )

    def _remember_fact(
        self,
        segment: Optional[str],
        user_agent: UserAgent,
        stack_snapshot: PreviousMessagesStack,
    ) -> Optional[str]:
        if not segment or not self._remember_policy.should_remember(segment):
            return None
        safe_segment = segment.replace('"', "") if segment else segment
        stack = self._clone_stack(stack_snapshot)
        return self._remember_command.execute(
            safe_segment,
            user_agent,
            stack,
        )

    def search(
        self,
        input_text: str,
        user_agent: Optional[UserAgent] = None,
        previous_messages_stack: Optional[PreviousMessagesStack] = None,
        synthesized_text_type: Optional[SynthesizedTextType] = None,
    ) -> Optional[str]:
        resolved_user_agent = self._resolve_user_agent(user_agent)
        stack = self._resolve_prev_message_stack(previous_messages_stack, resolved_user_agent)
        response = self._search_command.execute(
            input_text,
            resolved_user_agent,
            stack,
            synthesized_text_type=synthesized_text_type,
        )
        self._record_conversation_turn(resolved_user_agent, input_text, response)
        return response

    def _record_conversation_turn(
        self,
        user_agent: UserAgent,
        user_text: Optional[str],
        agent_response: Optional[str],
    ) -> None:
        if self._conversation_service is None:
            return
        if not user_text or not user_text.strip():
            return
        try:
            conversation = self._conversation_service.ensure_conversation_exists(
                user_id=user_agent.user_id,
                agent_id=user_agent.agent_id,
            )
            self._conversation_service.add_turn(
                conversation.id,
                user_agent.user_id,
                user_agent.agent_id,
                user_text,
                agent_response or "",
            )
        except Exception:
            try:
                app_logger.debug("[DeepthinkPipelineProcessor] Failed to persist conversation turn")
            except Exception:
                pass

    def generate_background_info(
        self,
        input_text: str,
        user_agent: Optional[UserAgent] = None,
        previous_messages_stack: Optional[PreviousMessagesStack] = None,
    ) -> str:
        resolved_user_agent = self._resolve_user_agent(user_agent)
        segments = self._large_text_processor.chunk_input(input_text)
        handler = _BackgroundInfoHandler(self._background_info_command, resolved_user_agent)
        contexts = self._fact_task_runner.run(
            segments,
            handler,
            mode=FactExecutionMode.SEQUENTIAL,
        )
        return "\n\n".join(context for context in contexts if context)

    def get_background_info(
        self,
        input_text: str,
        user_agent: Optional[UserAgent] = None,
        previous_messages_stack: Optional[PreviousMessagesStack] = None,
    ) -> str:
        return self.generate_background_info(
            input_text=input_text,
            user_agent=user_agent,
            previous_messages_stack=previous_messages_stack,
        )

    def _resolve_prev_message_stack(
        self,
        stack: Optional[PreviousMessagesStack],
        user_agent: UserAgent,
    ) -> PreviousMessagesStack:
        if stack is not None:
            return stack
        key = (user_agent.user_id, user_agent.agent_id)
        if key not in self._internal_stacks:
            self._internal_stacks[key] = PreviousMessagesStack()
        return self._internal_stacks[key]

    def _resolve_user_agent(
        self,
        user_agent: Optional[UserAgent],
    ) -> UserAgent:
        return self._user_agent_resolver.resolve_user_agent(user_agent)

    @staticmethod
    def _clone_stack(stack: PreviousMessagesStack) -> PreviousMessagesStack:
        cloned = PreviousMessagesStack()
        for role, text in stack.get_history_pairs():
            if role == "user":
                cloned.add_user(text)
            else:
                cloned.add_agent(text)
        return cloned

    def _synthesize_chat_response(
        self,
        responses: list[Optional[str]],
    ) -> Optional[str]:
        aggregated = _aggregate_responses(responses)
        if aggregated is None:
            return None
        return self._apply_chat_synthesizer(aggregated)

    def _apply_chat_synthesizer(self, text: str) -> str:
        if not text:
            return text
        if self._chat_synthesizer is None:
            return text
        return self._chat_synthesizer.synthesize(text)


def _aggregate_responses(responses: list[Optional[str]]) -> Optional[str]:
    parts = [item.strip() for item in responses if item and item.strip()]
    if not parts:
        return None
    aggregated = " ".join(parts).strip()
    return aggregated or None


def _default_remember_max_workers() -> int:
    cpu_count = resolve_cpu_count()
    computed = (cpu_count - 2) * 2
    return max(1, computed)


def _resolve_remember_max_workers() -> int:
    configured = app_config.get(
        "limits.remember_max_workers",
        _default_remember_max_workers(),
    )
    if configured <= 0:
        configured = _default_remember_max_workers()
    return max(1, configured)



from analogai.deepthink.application.usecases.learning.make_direct_statement.usecase import MakeDirectStatementUseCase
from analogai.deepthink.application.usecases.learning.make_direct_statement.models import MakeDirectStatementRequest, MakeDirectStatementResponse
from analogai.deepthink.application.usecases.learning.make_direct_statement.ports import LearnInputPort, MakeStatementOutputPort

__all__ = [
    "MakeDirectStatementUseCase",
    "MakeDirectStatementRequest",
    "MakeDirectStatementResponse",
    "LearnInputPort",
    "MakeStatementOutputPort",
]




"""Messaging helpers and orchestration utilities."""

from analogai.deepthink.messaging.fact.services.fact_service import FactServicePort
from analogai.deepthink.messaging.fact.services.fact_context_assembler import FactContextBuilder



from analogai.deepthink.thinker import DeepThinker
from analogai.deepthink.infrastructure.gateways.llm.enums.llm_provider import LlmProvider
from analogai.deepthink.application.value_objects import LlmConfig, EmbeddingConfig
from analogai.foundation.common.enums.embedding_type import EmbeddingType  # type: ignore[import-not-found]
from analogai.foundation.common.enums.embedding_provider import EmbeddingProvider  # type: ignore[import-not-found]


thinker = DeepThinker(
    llm_config=LlmConfig(
        provider=LlmProvider.AZURE,
        model="gpt-4o-mini",
    ),
    embedding_config=EmbeddingConfig(
        embed_type=EmbeddingType.API,
        provider=EmbeddingProvider.AZURE,
        model="text-embedding-3-small",
    ),
)


if __name__ == "__main__":
    while True:
        message = input("Message: ")
        result = thinker.chat(message)
        print(result if result else "(no response)")



from __future__ import annotations

import re
from typing import List


_SENTENCE_SPLIT_REGEX = re.compile(r"(?<=[.!?])\s+")


def split_into_sentences(text: str) -> List[str]:
    if text is None:
        raise ValueError("text cannot be None")
    cleaned = text.strip()
    if not cleaned:
        return []
    return [sentence for sentence in _SENTENCE_SPLIT_REGEX.split(cleaned) if sentence]


def chunk_sentences(text: str, max_sentences_per_chunk: int) -> List[str]:
    if text is None:
        raise ValueError("text cannot be None")
    cleaned = text.strip()
    if not cleaned:
        return []
    max_per_chunk = max(1, int(max_sentences_per_chunk))
    sentences = split_into_sentences(cleaned)
    chunks: List[str] = []
    current: List[str] = []
    for sentence in sentences:
        if not sentence:
            continue
        current.append(sentence)
        if len(current) >= max_per_chunk:
            chunks.append(" ".join(current).strip())
            current = []
    if current:
        chunks.append(" ".join(current).strip())
    return chunks



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import List, Optional

from analogai.deepthink.messaging.fact.fact import Fact


class FactRepositoryPort(ABC):
    @abstractmethod
    def register_turn(self, user_id: str, agent_id: str, turn_id: str) -> None:
        raise NotImplementedError

    @abstractmethod
    def save_fact(self, fact: Fact) -> None:
        raise NotImplementedError

    @abstractmethod
    def get_fact(self, id: str) -> Optional[Fact]:
        raise NotImplementedError

    @abstractmethod
    def update_facts_statement_id(self, turn_id: str, statement_id: str) -> None:
        raise NotImplementedError

    @abstractmethod
    def update_facts_notion_id(self, turn_id: str, notion_id: str) -> None:
        raise NotImplementedError

    @abstractmethod
    def update_facts_hypothetical_statement_id(self, turn_id: str, statement_id: str) -> None:
        raise NotImplementedError

    @abstractmethod
    def get_facts_for_turn(self, turn_id: str) -> List[Fact]:
        raise NotImplementedError

    @abstractmethod
    def get_turn_ids(self, user_id: str, agent_id: str) -> List[str]:
        raise NotImplementedError

    @abstractmethod
    def link_statement(self, statement_id: str, turn_id: str) -> None:
        raise NotImplementedError

    @abstractmethod
    def link_belief(self, belief_id: str, turn_id: str) -> None:
        raise NotImplementedError

    @abstractmethod
    def link_notion(self, notion_id: str, turn_id: str) -> None:
        raise NotImplementedError

    @abstractmethod
    def link_hypothetical_statement(self, statement_id: str, turn_id: str) -> None:
        raise NotImplementedError

    @abstractmethod
    def get_turn_id_for_statement(self, statement_id: str) -> Optional[str]:
        raise NotImplementedError

    @abstractmethod
    def get_turn_id_for_belief(self, belief_id: str) -> Optional[str]:
        raise NotImplementedError

    @abstractmethod
    def get_turn_id_for_notion(self, notion_id: str) -> Optional[str]:
        raise NotImplementedError

    @abstractmethod
    def get_turn_id_for_hypothetical_statement(self, statement_id: str) -> Optional[str]:
        raise NotImplementedError

    @abstractmethod
    def search_text(self, user_id: str, agent_id: str, query: str, top_k: int) -> List[Fact]:
        raise NotImplementedError

    @abstractmethod
    def search_similar(
        self,
        user_id: str,
        agent_id: str,
        embedding: List[float],
        top_k: int,
        similarity_cutoff: Optional[float] = None,
    ) -> List[Fact]:
        raise NotImplementedError



import re


_WEIGHT_UNITS = re.compile(r"\b(kg|g(?:rams?)?|lb|lbs|ounces?|oz)\b", re.IGNORECASE)
_AREA_UNITS = re.compile(
    r"\b(sq\s*\.?\s*m(?:eters?)?|m²|m\^2|square\s*m(?:eters?)?|sqm|sq\s*m|hectares?|ha|sq\s*ft|square\s*feet)\b",
    re.IGNORECASE,
)
_LENGTH_UNITS = re.compile(
    r"\b(m(?:eters?)?|cm|km|mm|feet|ft|inches?|in|yards?|yd|miles?|mi)\b",
    re.IGNORECASE,
)


PRICE_LIKE_TAGS = frozenset({"cost", "price"})

QUANTITY_LIKE_TAGS = frozenset({
    "quantity",
    "size",
    "weight",
    "height",
    "width",
    "length",
    "depth",
    "area",
    "distance",
})

_KEYWORD_TAGS = {
    "price": "price",
    "cost": "cost",
    "quantity": "quantity",
    "size": "size",
    "weight": "weight",
    "height": "height",
    "width": "width",
    "length": "length",
    "depth": "depth",
    "area": "area",
    "distance": "distance",
}


def detect_quantity_attribute_tag(caption: str):
    if not caption or not caption.strip():
        return None
    caption_text = caption.strip().lower()
    for key, tag in _KEYWORD_TAGS.items():
        if key in caption_text:
            return tag
    if _WEIGHT_UNITS.search(caption):
        return "weight"
    if _AREA_UNITS.search(caption):
        return "area"
    if _LENGTH_UNITS.search(caption):
        return "length"
    return None


def attribute_tag_from_tag(tag: str):
    if not tag or not tag.strip():
        return None
    key = tag.strip().lower()
    return _KEYWORD_TAGS.get(key, key)



from abc import ABC, abstractmethod
from typing import AsyncGenerator

class LlmGateway(ABC):
    @abstractmethod
    def generate_completion(
        self,
        user_message: str
    ) -> str:
        pass

    @abstractmethod
    async def generate_streaming_completion(
        self,
        user_message: str
    ) -> AsyncGenerator[str, None]:
        if False:
            yield ""

    @abstractmethod
    def set_system_prompt(self, system_prompt: str) -> None:
        pass



from analogai.deepthink.application.usecases.learning.make_direct_statement.models import MakeDirectStatementResponse
from analogai.deepthink.application.usecases.learning.make_direct_statement.ports import MakeStatementOutputPort
from analogai.foundation.common.enums import RelationshipType
from analogai.deepthink.presentation.presenters.presenter_response import PresenterResponse
from analogai.foundation.common.utils.expressions_serializer import serialize_belief_expression
from analogai.foundation.common.logging.app_logger import app_logger


class MakeDirectStatementPresenter(MakeStatementOutputPort):
    """
    Presenter for simple statement responses.
    Handles formatting of belief expressions for display.
    """
    
    def output_create_statement_success(self, response: MakeDirectStatementResponse):
        belief = response.belief_expression

        app_logger.info(f"Outputting learning success: {repr(belief)}")
        
        message = serialize_belief_expression(
            belief,
            timezone=response.user_agent.timezone,
        )
        
        return PresenterResponse(message=message)

    def output_statement_already_known(self, response: MakeDirectStatementResponse):
        belief = response.belief_expression

        app_logger.info("Outputting statement already known")
        serialized = serialize_belief_expression(belief)
        return PresenterResponse(message=f"{serialized}")

    def output_statement_update_success(self, response: MakeDirectStatementResponse):
        belief = response.belief_expression

        app_logger.info("Outputting statement update success")
        serialized = serialize_belief_expression(belief)
        return PresenterResponse(message=f"I see, will remember, that {serialized}")

    def output_update_not_applicable(self, response: MakeDirectStatementResponse):
        belief = response.belief_expression

        app_logger.info("Outputting update not applicable from learning response interface")
        serialized = serialize_belief_expression(belief)
        return PresenterResponse(message=f"I believe that {serialized}, will ask my companions about it")

    def output_forbidden(self, response: MakeDirectStatementResponse):
        app_logger.info("Outputting forbidden direct statement")
        return self.output_create_statement_success(response)




from abc import ABC, abstractmethod
from typing import Optional, Any
from analogai.deepthink.presentation.intents.intent import Intent
from analogai.deepthink.presentation.presenters.presenter_response import PresenterResponse


class IntentProcessingStrategy(ABC):
	@abstractmethod
	def execute(self, request: Any) -> tuple[Optional[PresenterResponse], Optional[Intent]]:
		pass


class ContextualIntentStrategy(IntentProcessingStrategy):
	def __init__(self, last_successful_intent: Intent):
		self._last_successful_intent = last_successful_intent
 
	def execute(self, request: Any) -> tuple[Optional[PresenterResponse], Optional[Intent]]:
		result = self._last_successful_intent.execute(request)
		return result if result is not None else None, None


class DefaultIntentStrategy(IntentProcessingStrategy):
	def __init__(self, intents: list[Intent]):
		self._intents = intents
 
	def execute(self, request: Any) -> tuple[Optional[PresenterResponse], Optional[Intent]]:
		for intent in self._intents:
			result = intent.execute(request)
			if result is not None:
				return result, intent
		return None, None






import unittest

from analogai.foundation.common.utils.expression_factory import ExpressionFactory


class TestNotionCaptionParser(unittest.TestCase):
    def setUp(self):
        self.factory = ExpressionFactory()



    def test_contextual_appositive(self):
        expression = self.factory.create_notion_expression("John, the programmer")
        if expression.caption.lower() == "john":
            self.assertEqual(len(expression.context), 1)
            self.assertEqual(expression.context[0].caption.lower(), "programmer")
        else:
            self.assertEqual(expression.caption.lower(), "john, the programmer")
            self.assertEqual(len(expression.context), 0)


    def test_article_head_is_stripped(self):
        expression = self.factory.create_notion_expression("the imperor")
        self.assertIn(expression.caption.lower(), {"imperor", "the imperor"})
        self.assertEqual(len(expression.context), 0)

    def test_alexander_the_great_without_comma_is_single_notion(self):
        expression = self.factory.create_notion_expression("alexander the great")
        self.assertEqual(expression.caption.lower(), "alexander the great")
        self.assertEqual(len(expression.context), 0)

    def test_alexander_comma_the_great_has_nested_context(self):
        expression = self.factory.create_notion_expression("alexander, the great")
        if expression.caption.lower() == "alexander":
            self.assertEqual(len(expression.context), 1)
            self.assertEqual(expression.context[0].caption.lower(), "great")
        else:
            self.assertEqual(expression.caption.lower(), "alexander, the great")
            self.assertEqual(len(expression.context), 0)

    def test_giorgi_comma_the_doctor_has_nested_context(self):
        expression = self.factory.create_notion_expression("Giorgi, the doctor")
        if expression.caption.lower() == "giorgi":
            self.assertEqual(len(expression.context), 1)
            self.assertEqual(expression.context[0].caption.lower(), "doctor")
        else:
            self.assertEqual(expression.caption.lower(), "giorgi, the doctor")
            self.assertEqual(len(expression.context), 0)

    def test_parses_or_as_single_caption(self):
        expression = self.factory.create_notion_expression("smoking or alcohol")
        self.assertEqual(expression.caption.lower(), "smoking or alcohol")
        self.assertEqual(len(expression.children), 0)

    def test_parses_and_as_single_caption(self):
        expression = self.factory.create_notion_expression("smoking and drinking")
        self.assertEqual(expression.caption.lower(), "smoking and drinking")
        self.assertEqual(len(expression.children), 0)

    def test_parses_or_with_three_items_as_single_caption(self):
        expression = self.factory.create_notion_expression("smoking or alcohol or drugs")
        self.assertEqual(expression.caption.lower(), "smoking or alcohol or drugs")
        self.assertEqual(len(expression.children), 0)

    def test_parses_and_with_contextual_as_single_caption(self):
        expression = self.factory.create_notion_expression("alexander, the great and socrates, the philosopher")
        self.assertEqual(expression.caption.lower(), "alexander, the great and socrates, the philosopher")
        self.assertEqual(len(expression.children), 0)

    def test_preserves_placeholder(self):
        expression = self.factory.create_notion_expression("respondent")
        self.assertEqual(expression.caption, "respondent")


if __name__ == "__main__":
    unittest.main()




from analogai.deepthink.application.value_objects.embedding_config import EmbeddingConfig
from analogai.deepthink.application.value_objects.llm_config import LlmConfig
from analogai.deepthink.application.value_objects.user_agent import UserAgent

__all__ = [
	"EmbeddingConfig",
	"LlmConfig",
	"UserAgent",
]



from __future__ import annotations

from typing import Iterable, List, Optional

from analogai.deepthink.messaging.fact.services.fact_context_service import FactContextWindowService
from analogai.foundation.common.logging.app_logger import app_logger
from analogai.deepthink.messaging.text.sentence_chunker import split_into_sentences


class AnswerFactAssembler:
    def __init__(self, fact_context_service: Optional[FactContextWindowService] = None) -> None:
        self._fact_context_service = fact_context_service

    def build_fact_block(
        self,
        beliefs: Iterable,
        belief_lines: Iterable[str],
        notion_lines: Iterable[str],
        fact_lookup: dict[str, str],
        raw_answer: str,
        qa_history: str,
        user_history_block: str,
        include_raw_answer: bool = True,
    ) -> str:
        items: List[str] = []
        if fact_lookup:
            for fact_key, fact_text in fact_lookup.items():
                cleaned = _sanitize_fact_line(fact_text)
                if not cleaned:
                    continue
                try:
                    app_logger.debug(
                        "[AnswerFactAssembler] fact added to facts id=%s value=%s",
                        fact_key,
                        cleaned,
                    )
                except Exception:
                    pass
                _extend_fact_items(items, cleaned)
        if not items:
            return ""
        return "\n".join(f"{index}) {item}" for index, item in enumerate(items, start=1))


def _looks_like_question(text: str) -> bool:
    if "?" in text:
        return True
    lowered = text.strip().lower()
    return lowered.startswith("q:") or lowered.startswith("question:")


def _sanitize_fact_line(text: str) -> str:
    if not text:
        return ""
    cleaned = str(text).strip()
    if "A:" in cleaned:
        cleaned = cleaned.rsplit("A:", 1)[-1].strip()
        cleaned = cleaned.lstrip("|-–— ")
    if not cleaned:
        return ""
    if _looks_like_question(cleaned):
        return ""
    return cleaned


def _normalize_fact_line(text: str) -> str:
    normalized = text.strip().lower()
    if normalized.startswith("-"):
        normalized = normalized.lstrip("- ")
    return normalized


def _append_fact_item(items: List[str], item: str) -> None:
    items.append(item)


def _extend_fact_items(items: List[str], text: str) -> None:
    sentences = split_into_sentences(text)
    if not sentences:
        _append_fact_item(items, text)
        return
    for sentence in sentences:
        cleaned = sentence.strip()
        if not cleaned:
            continue
        _append_fact_item(items, cleaned)



from __future__ import annotations

from dataclasses import dataclass
from typing import Optional, Protocol

from analogai.deepthink.messaging.text.sentence_chunker import (
    chunk_sentences,
    split_into_sentences,
)


class ChunkHandler(Protocol):
    def handle(self, chunk: str) -> Optional[str]:
        ...


@dataclass(frozen=True)
class SentenceBatcher:
    max_sentences_per_input: int

    def chunk_input(self, input_text: str) -> list[str]:
        return chunk_sentences(input_text, self.max_sentences_per_input)

    def process_chat(
        self,
        input_text: str,
        handler: ChunkHandler,
    ) -> Optional[str]:
        chunks = self.chunk_input(input_text)
        if not chunks:
            return None
        last_response: Optional[str] = None
        for chunk in chunks:
            last_response = handler.handle(chunk)
        if not self._last_sentence_is_question(input_text):
            if last_response is not None:
                return last_response
            if self._is_single_sentence(input_text):
                return input_text.strip()
            return input_text.strip() or None
        return last_response

    def _last_sentence_is_question(self, input_text: str) -> bool:
        sentences = split_into_sentences(input_text)
        if not sentences:
            return False
        last_sentence = sentences[-1].strip()
        if not last_sentence:
            return False
        return "?" in last_sentence

    def _is_single_sentence(self, input_text: str) -> bool:
        sentences = split_into_sentences(input_text)
        return len(sentences) == 1



from __future__ import annotations

import json
import re
from typing import Any


class JsonFenceCleaner:
    def clean(self, text: str) -> str:
        cleaned = text.strip()
        if not cleaned:
            return cleaned
        if cleaned.startswith("```"):
            cleaned = re.sub(r"^```[a-zA-Z]*\n?", "", cleaned)
            cleaned = re.sub(r"```$", "", cleaned).strip()
        return cleaned


class SentenceArrayRepairer:
    def repair(self, text: str) -> str:
        if '"sentences"' not in text:
            return text
        pattern = re.compile(r"(\]\}\}|\}\})\s*,\s*\{\"intent_type\"")

        def _insert_missing_brace(match: re.Match) -> str:
            return match.group(1) + "},{\"intent_type\""

        return pattern.sub(_insert_missing_brace, text)


class JsonObjectExtractor:
    def extract(self, text: str) -> Any:
        decoder = json.JSONDecoder()
        index = 0
        length = len(text)
        objects = []

        while index < length:
            next_obj = text.find("{", index)
            next_arr = text.find("[", index)
            if next_obj == -1 and next_arr == -1:
                break
            if next_obj == -1 or (next_arr != -1 and next_arr < next_obj):
                index = next_arr
            else:
                index = next_obj

            try:
                obj, end = decoder.raw_decode(text[index:])
            except json.JSONDecodeError:
                break

            objects.append(obj)
            index += end

            while index < length and text[index] in " \t\r\n,":
                index += 1

        if not objects:
            return None
        if len(objects) == 1:
            return objects[0]
        return objects


class TransformedSentenceParser:
    def __init__(
        self,
        fence_cleaner: JsonFenceCleaner | None = None,
        repairer: SentenceArrayRepairer | None = None,
        extractor: JsonObjectExtractor | None = None,
    ) -> None:
        self._fence_cleaner = fence_cleaner or JsonFenceCleaner()
        self._repairer = repairer or SentenceArrayRepairer()
        self._extractor = extractor or JsonObjectExtractor()

    def parse(self, text: str) -> Any:
        cleaned = self._fence_cleaner.clean(text)
        if not cleaned:
            raise json.JSONDecodeError("Empty JSON", cleaned, 0)

        try:
            return json.loads(cleaned)
        except json.JSONDecodeError:
            repaired = self._repairer.repair(cleaned)
            if repaired != cleaned:
                try:
                    return json.loads(repaired)
                except json.JSONDecodeError:
                    pass

        extracted = self._extractor.extract(cleaned)
        if extracted is not None:
            return extracted

        return json.loads(cleaned)

    def format_for_log(self, text: str) -> str:
        if text is None:
            return "(none)"
        cleaned = text.strip()
        if not cleaned:
            return "(empty)"
        try:
            parsed = self.parse(cleaned)
            return json.dumps(parsed, indent=2, ensure_ascii=False)
        except json.JSONDecodeError:
            return cleaned



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Optional

from analogai.deepthink.application.value_objects.user_agent import UserAgent
from analogai.deepthink.messaging.question_answering.question_answering_service import (
    QuestionAnsweringService,
)
from analogai.deepthink.presentation.pipeline.pipeline import MessagePipeline
from analogai.deepthink.presentation.preprocessing.previous_messages_stack import PreviousMessagesStack
from analogai.deepthink.messaging.question_answering.enums.synthesized_text_type import (
    SynthesizedTextType,
)


class MessageCommand(ABC):
    @abstractmethod
    def execute(
        self,
        text: str,
        user_agent: UserAgent,
        stack: PreviousMessagesStack,
        synthesized_text_type: Optional[SynthesizedTextType] = None,
    ) -> Optional[str]:
        raise NotImplementedError


class RememberCommand(MessageCommand):
    def __init__(self, pipeline: MessagePipeline):
        self._pipeline = pipeline

    def execute(
        self,
        text: str,
        user_agent: UserAgent,
        stack: PreviousMessagesStack,
        synthesized_text_type: Optional[SynthesizedTextType] = None,
    ) -> Optional[str]:
        stack.add_user(text)
        result = self._pipeline.execute(
            raw_message=text,
            user_agent=user_agent,
            previous_messages_stack=stack,
        )
        if result is None:
            message_text = None
        else:
            try:
                message_text = result.message_text
            except AttributeError:
                message_text = None
        if message_text:
            stack.add_agent(message_text)
        return message_text


class SearchCommand(MessageCommand):
    def __init__(self, question_answering_service: QuestionAnsweringService):
        self._question_answering_service = question_answering_service

    def execute(
        self,
        text: str,
        user_agent: UserAgent,
        stack: PreviousMessagesStack,
        synthesized_text_type: Optional[SynthesizedTextType] = None,
    ) -> Optional[str]:
        stack.add_user(text)
        answer = self._question_answering_service.answer_question(
            question_text=text,
            user_agent=user_agent,
            synthesized_text_type=synthesized_text_type,
            previous_messages_stack=stack,
        )
        if answer:
            stack.add_agent(answer)
        return answer


class GenerateBackgroundInfoCommand:
    def __init__(self, question_answering_service: QuestionAnsweringService):
        self._question_answering_service = question_answering_service

    def execute(self, text: str, user_agent: UserAgent) -> str:
        return self._question_answering_service.build_question_background_info(
            question_text=text,
            user_agent=user_agent,
        )


__all__ = [
    "GenerateBackgroundInfoCommand",
    "MessageCommand",
    "RememberCommand",
    "SearchCommand",
]



def times_match(time1, time2) -> bool:
    if time1 is None and time2 is None:
        return True
    if time1 is None or time2 is None:
        return False
    return (
        time1.year == time2.year
        and time1.month == time2.month
        and time1.day == time2.day
        and time1.hour == time2.hour
        and time1.minute == time2.minute
    )


def modifiers_match(modal1, modal2) -> bool:
    if modal1 is None and modal2 is None:
        return True
    if modal1 is None or modal2 is None:
        return False
    if modal1.location != modal2.location:
        return False
    if modal1.deontic_modality != modal2.deontic_modality:
        return False
    if not times_match(modal1.from_time, modal2.from_time):
        return False
    if not times_match(modal1.to_time, modal2.to_time):
        return False
    return True



from dataclasses import dataclass
from typing import Optional, Any

from analogai.deepthink.presentation.intent_type import IntentType
# from analogai.deepthink.presentation.user_message import UserMessage
from analogai.deepthink.presentation.pipeline.results import (
    PreprocessResult,
    TransformResult,
    ParseResult,
    InterpretResult,
    IntentResult,
    PostprocessResult,
    OutputResult,
)


@dataclass
class PipelineState:
    # raw_message: str
    user_agent: Optional[Any] = None
    message_text: Optional[str] = None
    intent_type: Optional[IntentType] = None
    request: Optional[Any] = None
    preprocess_result: Optional[PreprocessResult] = None
    transform_result: Optional[TransformResult] = None
    parse_result: Optional[ParseResult] = None
    interpret_result: Optional[InterpretResult] = None
    intent_result: Optional[IntentResult] = None
    postprocess_result: Optional[PostprocessResult] = None
    output_result: Optional[OutputResult] = None



from __future__ import annotations

from typing import Optional

from analogai.deepthink.messaging.question_answering.question_answering_service import (
    QuestionAnsweringService,
)
from analogai.deepthink.messaging.commands import GenerateBackgroundInfoCommand, RememberCommand, SearchCommand
from analogai.deepthink.presentation.pipeline.pipeline import MessagePipeline


class ChatCommandFactory:
    def __init__(
        self,
        pipeline: MessagePipeline,
        question_answering_service: QuestionAnsweringService,
    ):
        self._pipeline = pipeline
        self._question_answering_service = question_answering_service

    def create_remember_command(self) -> RememberCommand:
        return RememberCommand(self._pipeline)

    def create_search_command(self) -> SearchCommand:
        return SearchCommand(self._question_answering_service)

    def create_background_info_command(self) -> GenerateBackgroundInfoCommand:
        return GenerateBackgroundInfoCommand(self._question_answering_service)



import unittest
from analogai.deepthink.presentation.preprocessing.preprocessors.pronoun_preprocessor import (
    preprocess_pronouns,
)
from analogai.deepthink.presentation.enums.pronoun_placeholder import PronounPlaceholder


class TestPronounPreprocessor(unittest.TestCase):
    def test_preprocess_pronouns_you_to_agent_placeholder(self):
        result = preprocess_pronouns("you are great")
        self.assertEqual(result, f"{PronounPlaceholder.AGENT_PLACEHOLDER.value} is great")

    def test_preprocess_pronouns_your_to_agent_placeholder_s(self):
        result = preprocess_pronouns("your voice sounds good")
        self.assertEqual(result, f"{PronounPlaceholder.AGENT_PLACEHOLDER.value}'s voice sounds good")

    def test_preprocess_pronouns_we_becomes_company_placeholder(self):
        result = preprocess_pronouns("we sell iphones")
        self.assertEqual(result, f"{PronounPlaceholder.COMPANY_PLACEHOLDER.value} sell iphones")

    def test_preprocess_pronouns_our_becomes_company_placeholder(self):
        result = preprocess_pronouns("our products are good")
        self.assertEqual(result, f"{PronounPlaceholder.COMPANY_PLACEHOLDER.value}'s products are good")

    def test_preprocess_pronouns_i_to_respondent_placeholder(self):
        result = preprocess_pronouns("i am here")
        self.assertEqual(result, f"{PronounPlaceholder.RESPONDENT_PLACEHOLDER.value} is here")

    def test_preprocess_pronouns_am_respondent_to_is_respondent(self):
        respondent = PronounPlaceholder.RESPONDENT_PLACEHOLDER.value
        result = preprocess_pronouns(f"am {respondent} here")
        self.assertEqual(result, f"is {respondent} here")

    def test_preprocess_pronouns_agent_are_to_agent_is(self):
        agent = PronounPlaceholder.AGENT_PLACEHOLDER.value
        result = preprocess_pronouns(f"{agent} are nice")
        self.assertEqual(result, f"{agent} is nice")

    def test_preprocess_pronouns_are_agent_to_is_agent(self):
        agent = PronounPlaceholder.AGENT_PLACEHOLDER.value
        result = preprocess_pronouns(f"are {agent} nice")
        self.assertEqual(result, f"is {agent} nice")

    def test_preprocess_pronouns_my_to_respondent_placeholder_s(self):
        result = preprocess_pronouns("my name is john")
        self.assertEqual(result, f"{PronounPlaceholder.RESPONDENT_PLACEHOLDER.value}'s name is john")

    def test_preprocess_pronouns_complex_sentence(self):
        result = preprocess_pronouns("i think you are great and we should work together")
        agent = PronounPlaceholder.AGENT_PLACEHOLDER.value
        company = PronounPlaceholder.COMPANY_PLACEHOLDER.value
        respondent = PronounPlaceholder.RESPONDENT_PLACEHOLDER.value
        self.assertEqual(
            result,
            f"{respondent} think {agent} is great and {company} should work together"
        )

    def test_preprocess_pronouns_from_pronoun_you_capitalized(self):
        result = preprocess_pronouns("You are great")
        self.assertEqual(result, f"{PronounPlaceholder.AGENT_PLACEHOLDER.value} is great")

    def test_preprocess_pronouns_from_pronoun_i_capitalized(self):
        result = preprocess_pronouns("I think so")
        self.assertEqual(result, f"{PronounPlaceholder.RESPONDENT_PLACEHOLDER.value} think so")


if __name__ == "__main__":
    unittest.main()


from __future__ import annotations

from typing import Iterable, List

from analogai.foundation.common.dtos.notion_expression import NotionExpression  # type: ignore[import-not-found]

from analogai.deepthink.application.value_objects.user_agent import UserAgent


class BeliefSearchAdapter:
    def __init__(
        self,
        belief_service,
        max_entity_depth: int = 5,
    ) -> None:
        self._belief_service = belief_service
        self._max_entity_depth = max(0, int(max_entity_depth))

    def search_beliefs(self, user_agent: UserAgent, entities: Iterable[str]) -> List:
        if self._belief_service is None:
            return []
        results: List = []
        seen_belief_ids = set()
        entity_queue = [(entity, 0) for entity in entities if entity]
        seen_entity_keys = {entity.lower() for entity, _ in entity_queue}

        while entity_queue:
            entity, depth = entity_queue.pop(0)
            if depth > self._max_entity_depth:
                continue
            notion_expression = NotionExpression(caption=entity)
            subject_beliefs = (
                self._belief_service.search_by_subject_notion_expression(
                    user_agent.agent_id,
                    notion_expression,
                )
                or []
            )
            complement_beliefs = (
                self._belief_service.search_by_complement_notion_expression(
                    user_agent.agent_id,
                    notion_expression,
                )
                or []
            )
            for belief in [*subject_beliefs, *complement_beliefs]:
                belief_id = belief.id
                if belief_id in seen_belief_ids:
                    continue
                seen_belief_ids.add(belief_id)
                results.append(belief)

                if depth < self._max_entity_depth:
                    for related in _extract_belief_entities(belief):
                        related_key = related.lower()
                        if related_key in seen_entity_keys:
                            continue
                        seen_entity_keys.add(related_key)
                        entity_queue.append((related, depth + 1))

        return results


def _extract_belief_entities(belief) -> List[str]:
    entities: List[str] = []
    subject_notion, complement_notion = _belief_notions(belief)
    subject_caption = _notion_caption(subject_notion)
    complement_caption = _notion_caption(complement_notion)
    if subject_caption:
        entities.append(str(subject_caption))
    if complement_caption:
        entities.append(str(complement_caption))
    return entities


def _belief_notions(belief) -> tuple[object | None, object | None]:
    if belief is None:
        return (None, None)
    return (belief.subject_notion, belief.complement_notion)


def _notion_caption(notion) -> str | None:
    if notion is None:
        return None
    return notion.caption



from __future__ import annotations

from abc import ABC, abstractmethod
from datetime import datetime
from typing import Dict, Iterable, List, Optional, Tuple


class NotionEnrichmentSleepRepository(ABC):
    @abstractmethod
    def get_last_sleep_at(self, user_id: str, agent_id: str) -> Optional[datetime]:
        raise NotImplementedError

    @abstractmethod
    def set_last_sleep_at(self, user_id: str, agent_id: str, sleep_time: datetime) -> None:
        raise NotImplementedError

    @abstractmethod
    def list_unenriched_notions_created_after(
        self,
        user_id: str,
        agent_id: str,
        since: Optional[datetime],
        limit: Optional[int] = None,
    ) -> List[object]:
        raise NotImplementedError

    @abstractmethod
    def mark_enriched(self, user_id: str, agent_id: str, notion: object) -> None:
        raise NotImplementedError


class InMemoryNotionEnrichmentSleepRepository(NotionEnrichmentSleepRepository):
    def __init__(self):
        self._last_sleep: Dict[Tuple[str, str], datetime] = {}
        self._notions: Dict[Tuple[str, str], List[object]] = {}
        self._enriched_ids: set[str] = set()

    def get_last_sleep_at(self, user_id: str, agent_id: str) -> Optional[datetime]:
        return self._last_sleep.get((user_id, agent_id))

    def set_last_sleep_at(self, user_id: str, agent_id: str, sleep_time: datetime) -> None:
        self._last_sleep[(user_id, agent_id)] = sleep_time

    def list_unenriched_notions_created_after(
        self,
        user_id: str,
        agent_id: str,
        since: Optional[datetime],
        limit: Optional[int] = None,
    ) -> List[object]:
        key = (user_id, agent_id)
        notions = self._notions.get(key, [])
        filtered: List[object] = []
        for notion in notions:
            if self._is_enriched(notion):
                continue
            created_at = _get_notion_created_at(notion)
            if since is not None and created_at is not None and created_at <= since:
                continue
            filtered.append(notion)
            if limit is not None and len(filtered) >= limit:
                break
        return filtered

    def mark_enriched(self, user_id: str, agent_id: str, notion: object) -> None:
        notion_id = _get_notion_id(notion)
        if notion_id:
            self._enriched_ids.add(notion_id)

    def seed_notions(self, user_id: str, agent_id: str, notions: Iterable[object]) -> None:
        key = (user_id, agent_id)
        self._notions.setdefault(key, []).extend(list(notions))

    def _is_enriched(self, notion: object) -> bool:
        notion_id = _get_notion_id(notion)
        if notion_id and notion_id in self._enriched_ids:
            return True
        try:
            return bool(notion.is_enriched)
        except AttributeError:
            return False


def _get_notion_id(notion: object) -> Optional[str]:
    if notion is None:
        return None
    try:
        notion_id = notion.id
    except AttributeError:
        return None
    if notion_id is None:
        return None
    return str(notion_id)


def _get_notion_created_at(notion: object) -> Optional[datetime]:
    if notion is None:
        return None
    try:
        value = notion.created_at
        if value is not None:
            return value
    except AttributeError:
        pass
    try:
        value = notion.createdAt
        if value is not None:
            return value
    except AttributeError:
        pass
    try:
        value = notion.created_on
        if value is not None:
            return value
    except AttributeError:
        pass
    try:
        value = notion.createdOn
        if value is not None:
            return value
    except AttributeError:
        pass
    return None



from __future__ import annotations

from typing import Optional, Protocol

from analogai.deepthink.application.value_objects.user_agent import UserAgent
from analogai.deepthink.presentation.preprocessing.previous_messages_stack import PreviousMessagesStack
from analogai.deepthink.messaging.question_answering.question_answering_service import (
    QuestionAnsweringService,
)
from analogai.deepthink.messaging.question_answering.question_mark_detector import detect_question_mark


class MessageRouteHandler(Protocol):
    def can_handle(self, text: str) -> bool:
        ...


class StatementHandler(Protocol):
    def handle(
        self,
        text: str,
        user_agent: UserAgent,
        previous_messages_stack: PreviousMessagesStack,
    ) -> Optional[str]:
        ...


class QuestionAnswerHandler:
    def __init__(self, question_service: QuestionAnsweringService):
        self._question_service = question_service

    def can_handle(self, text: str) -> bool:
        return detect_question_mark(text)

    def handle(
        self,
        text: str,
        user_agent: UserAgent,
        previous_messages_stack: PreviousMessagesStack,
    ) -> Optional[str]:
        return self._question_service.answer_question(
            question_text=text,
            user_agent=user_agent,
            previous_messages_stack=previous_messages_stack,
        )


class StatementFallbackHandler:
    def __init__(
        self,
        handler: StatementHandler,
    ):
        self._handler = handler

    def can_handle(self, text: str) -> bool:
        return True

    def handle(
        self,
        text: str,
        user_agent: UserAgent,
        previous_messages_stack: PreviousMessagesStack,
    ) -> Optional[str]:
        return self._handler.handle(text, user_agent, previous_messages_stack)


class QuestionRoutingService:
    def __init__(self, question_service: QuestionAnsweringService):
        self._question_service = question_service
        self._question_handler = QuestionAnswerHandler(question_service)

    def is_question_message(self, text: str) -> bool:
        return self._question_handler.can_handle(text)

    def route_message(
        self,
        raw_message: str,
        user_agent: UserAgent,
        previous_messages_stack: PreviousMessagesStack,
        statement_handler: StatementHandler,
    ) -> Optional[str]:
        handlers: list[MessageRouteHandler] = [
            self._question_handler,
            StatementFallbackHandler(statement_handler),
        ]
        for handler in handlers:
            if handler.can_handle(raw_message):
                return handler.handle(raw_message, user_agent, previous_messages_stack)
        return None



from collections import deque
from typing import Sequence, TYPE_CHECKING

from analogai.deepthink.presentation.preprocessing.previous_messages_stack import PreviousMessagesStack

if TYPE_CHECKING:
    from analogai.foundation.extensions.conversation.services.conversation_service import ConversationService


class ConversationPreviousMessagesStackAdapter(PreviousMessagesStack):
    MAX_HISTORY_LIMIT = 5

    def __init__(
        self,
        conversation_service: "ConversationService",
        user_id: str,
        agent_id: str,
        history_limit: int = 5,
    ):
        super().__init__(maxlen=min(history_limit, self.MAX_HISTORY_LIMIT))
        self._conversation_service = conversation_service
        self._user_id = user_id
        self._agent_id = agent_id
        self._history_limit = min(history_limit, self.MAX_HISTORY_LIMIT)
        self._cached_conversation = None

    def add(self, text: str) -> None:
        pass

    def get_history(self) -> Sequence[str]:
        conversation = self._cached_conversation
        if conversation is None:
            conversation = self._conversation_service.get_latest_conversation(
                self._user_id, self._agent_id
            )
            self._cached_conversation = conversation
        if not conversation:
            return ()
        history = deque(maxlen=self.MAX_HISTORY_LIMIT)
        for turn in conversation.get_recent_turns(limit=self._history_limit):
            for part in (turn.user_message, turn.agent_response):
                if part and (cleaned := part.strip()):
                    history.append(cleaned)
        return tuple(history)

    def refresh(self) -> None:
        self._cached_conversation = None



import unittest
from collections import deque
from unittest import TestCase
from unittest.mock import MagicMock
from analogai.deepthink.application.usecases.chat.chat_response_stack import ChatResponseStack
from analogai.deepthink.tests.mocks.entities.mock_belief_factory import MockBeliefFactory
from analogai.deepthink.tests.mocks.entities.mock_notion_factory import MockNotionFactory

class TestChatResponseStack(TestCase):
    def setUp(self):
        self.stack = ChatResponseStack()
        self.user_id = "test-user"
        self.agent_id = "test-agent"

        def create_belief(object_caption: str, relation: str, subject_caption: str):
            subject_notion = MockNotionFactory.create(caption=subject_caption)
            complement_notion = MockNotionFactory.create(caption=object_caption)
            belief = MockBeliefFactory.create(
                subject_notion=subject_notion,
                complement_notion=complement_notion,
            )
            belief.relation = relation
            return belief

        self.create_belief = create_belief

    def test_add_valid_response(self):
        response = self.create_belief("test_response", "positive", "user")
        self.stack.add(response)
        self.assertEqual(self.stack.get_size(), 1)
        latest = self.stack.get_latest()
        self.assertEqual(latest.complement_notion.caption, response.complement_notion.caption)
        self.assertEqual(latest.relation, response.relation)
        self.assertEqual(latest.subject_notion.caption, response.subject_notion.caption)

    def test_add_empty_response(self):
        initial_size = self.stack.get_size()
        self.stack.add(self.create_belief("", "positive", "user"))
        self.assertEqual(self.stack.get_size(), initial_size + 1)

    def test_add_none_response(self):
        initial_size = self.stack.get_size()
        self.stack.add(None)
        self.assertEqual(self.stack.get_size(), initial_size)

    def test_add_exceeds_max_size(self):
        for i in range(ChatResponseStack.MAX_SIZE + 2):
            self.stack.add(self.create_belief(f"response_{i}", "positive", "user"))
        self.assertEqual(self.stack.get_size(), ChatResponseStack.MAX_SIZE)
        expected = [
            self.create_belief(f"response_{i}", "positive", "user")
            for i in range(2, ChatResponseStack.MAX_SIZE + 2)
        ]
        actual = list(self.stack.get_all())
        for a, e in zip(actual, expected):
            self.assertEqual(a.complement_notion.caption, e.complement_notion.caption)
            self.assertEqual(a.relation, e.relation)
            self.assertEqual(a.subject_notion.caption, e.subject_notion.caption)

    def test_get_latest_empty_stack(self):
        result = self.stack.get_latest()
        self.assertIsNone(result)

    def test_get_latest_with_responses(self):
        self.stack.add(self.create_belief("first", "positive", "user"))
        self.stack.add(self.create_belief("second", "positive", "user"))
        result = self.stack.get_latest()
        self.assertEqual(result.complement_notion.caption, "second")
        self.assertEqual(result.relation, "positive")
        self.assertEqual(result.subject_notion.caption, "user")

    def test_get_size_empty(self):
        self.assertEqual(self.stack.get_size(), 0)

    def test_get_size_with_responses(self):
        self.stack.add(self.create_belief("first", "positive", "user"))
        self.stack.add(self.create_belief("second", "positive", "user"))
        self.assertEqual(self.stack.get_size(), 2)

    def test_get_all_empty(self):
        self.assertEqual(list(self.stack.get_all()), [])

    def test_get_all_with_responses(self):
        responses = [
            self.create_belief("first", "positive", "user"),
            self.create_belief("second", "positive", "user"),
            self.create_belief("third", "positive", "user")
        ]
        for response in responses:
            self.stack.add(response)
        actual = list(self.stack.get_all())
        for a, e in zip(actual, responses):
            self.assertEqual(a.complement_notion.caption, e.complement_notion.caption)
            self.assertEqual(a.relation, e.relation)
            self.assertEqual(a.subject_notion.caption, e.subject_notion.caption)

    def test_clear(self):
        self.stack.add(self.create_belief("first", "positive", "user"))
        self.stack.add(self.create_belief("second", "positive", "user"))
        self.stack.clear()
        self.assertEqual(self.stack.get_size(), 0)
        self.assertEqual(list(self.stack.get_all()), [])
        result = self.stack.get_latest()
        self.assertIsNone(result)

if __name__ == '__main__':
    unittest.main()





from __future__ import annotations

from typing import Optional, List

from analogai.deepthink.messaging.fact.services.fact_service import FactServicePort
from analogai.deepthink.messaging.fact.utils.fact_utils import (
    collect_fact_window,
    format_fact_summary,
    get_belief_id,
    get_statement_id,
)


class FactContextBuilder:
    def __init__(self, fact_service: FactServicePort):
        self._fact_service = fact_service

    def build_belief_context(self, belief) -> Optional[str]:
        resolved_context = self._resolve_facts(belief)
        if not resolved_context:
            return None
        _, fact_list = resolved_context
        if not fact_list:
            return None
        anchor_fact = fact_list[0]
        summaries: List[str] = [
            format_fact_summary(fact)
            for fact in fact_list
            if fact.text
        ]
        window_facts = collect_fact_window(anchor_fact, window=5)
        for neighbor_fact in window_facts:
            summaries.append(format_fact_summary(neighbor_fact))
        filtered_summaries = [summary for summary in summaries if summary]
        return "; ".join(filtered_summaries) if filtered_summaries else None

    def _resolve_facts(self, belief):
        belief_id = get_belief_id(belief)
        if belief_id:
            fact_list = self._fact_service.get_facts_for_belief(belief_id)
            if fact_list:
                turn_id = fact_list[0].turn_id
                if turn_id:
                    return turn_id, fact_list
        statement_id = get_statement_id(belief)
        if statement_id:
            fact_list = self._fact_service.get_facts_for_statement(statement_id)
            if fact_list:
                turn_id = fact_list[0].turn_id
                if turn_id:
                    return turn_id, fact_list
        return None



from typing import Any

from analogai.deepthink.presentation.interpreters.interpreter_result import InterpreterResult


class Interpreter:
    def interpret(
        self,
        parsed_message: Any,
        user_agent: Any,
        message_text: str,
    ) -> InterpreterResult | None | bool:
        raise NotImplementedError("Subclasses must implement this method")




from typing import AsyncGenerator
import anthropic
from analogai.foundation.settings import config as infra_config
from analogai.deepthink.infrastructure.gateways.llm.gateway import LlmGateway


class AnthropicLlmGatewayImpl(LlmGateway):
    def __init__(
        self,
        system_prompt: str | None = None,
        max_tokens: int = 300,
        model: str | None = None,
    ):
        api_key = infra_config.get("ANTHROPIC_API_KEY", "")
        if not api_key:
            raise ValueError("ANTHROPIC_API_KEY must be set for Anthropic gateway")
        if not model:
            raise ValueError("Model name must be provided for Anthropic gateway")
        self.model = model
        self.default_system_prompt = system_prompt or ""
        self.default_max_tokens = min(max_tokens, 300)
        self.client = anthropic.Anthropic(api_key=api_key)

    def set_model(self, model: str) -> None:
        self.model = model

    def set_system_prompt(self, system_prompt: str) -> None:
        self.default_system_prompt = system_prompt

    def generate_completion(
        self,
        user_message: str,
    ) -> str:
        max_tokens = self.default_max_tokens
        final_system_prompt = self.default_system_prompt
        response = self.client.messages.create(
            model=self.model,
            max_tokens=max_tokens,
            system=final_system_prompt,
            messages=[{"role": "user", "content": user_message}],
        )
        parts = []
        for block in response.content or []:
            try:
                text = block.text
            except AttributeError:
                text = None
            if text:
                parts.append(text)
        return "".join(parts).strip()

    async def generate_streaming_completion(
        self,
        user_message: str,
    ) -> AsyncGenerator[str, None]:
        max_tokens = self.default_max_tokens
        final_system_prompt = self.default_system_prompt
        stream = self.client.messages.stream(
            model=self.model,
            max_tokens=max_tokens,
            system=final_system_prompt,
            messages=[{"role": "user", "content": user_message}],
        )

        with stream as response_stream:
            for event in response_stream:
                try:
                    event_type = event.type
                except AttributeError:
                    event_type = None
                if event_type != "content_block_delta":
                    continue
                try:
                    delta = event.delta
                except AttributeError:
                    delta = None
                if delta is None:
                    continue
                try:
                    content = delta.text
                except AttributeError:
                    content = None
                if content:
                    yield content


if __name__ == "__main__":
    gateway = AnthropicLlmGatewayImpl(model="claude-sonnet-4-6")
    result = gateway.generate_completion("Hello, how are you?")
    print(result)


from dataclasses import dataclass
from typing import List
from analogai.foundation.common.enums import RelationshipType


@dataclass
class EnrichmentSuggestion:
    subject_caption: str
    relationship_type: RelationshipType
    complement_caption: str


@dataclass
class EnrichmentResult:
    notion_caption: str
    suggestions: List[EnrichmentSuggestion]



from __future__ import annotations


_NO_RESULT = object()


class InterpreterResultResolver:
    def resolve(self, result):
        if result is False:
            return False
        if result is not None:
            return result
        return _NO_RESULT

    def is_resolved(self, value) -> bool:
        return value is not _NO_RESULT

    @property
    def no_result(self):
        return _NO_RESULT



import json
import re
from pathlib import Path

from analogai.deepthink.infrastructure.gateways.llm.llm_factory import LlmGatewayFactory
from analogai.deepthink.infrastructure.gateways.llm.providers.sentence_transformer_examples import (
    get_enhanced_system_prompt,
)
from analogai.foundation.common.logging.app_logger import app_logger
from analogai.deepthink.presentation.preprocessing.previous_messages_stack import PreviousMessagesStack
from analogai.deepthink.presentation.parsers.json_sentence_parser import is_json_format
from analogai.deepthink.presentation.preprocessing.preprocessors.llm.transformed_sentence_normalizer import (
    normalize_transformed_sentence,
)
from analogai.deepthink.presentation.preprocessing.preprocessors.llm.transformed_sentence_parser import (
    TransformedSentenceParser,
)


def get_sentence_transformer():
    return transform_sentence_with_llm


def _load_sentence_transformer_prompt() -> str:
    prompts_dir = Path(__file__).resolve().parents[3] / "infrastructure" / "prompts"
    prompt_path = prompts_dir / "sentence_transformer_prompt.txt"
    if not prompt_path.exists():
        raise FileNotFoundError(f"Prompt file not found: {prompt_path}")
    base_prompt = prompt_path.read_text(encoding="utf-8").strip()
    return get_enhanced_system_prompt(base_prompt)


def transform_sentence_with_llm(text: str, previous_messages_stack: PreviousMessagesStack) -> str:
    if not text or not text.strip():
        return text

    parser = TransformedSentenceParser()

    if is_json_format(text):
        app_logger.info(f"[SentenceTransformer] Skipping LLM - already JSON format")
        try:
            data = parser.parse(text.strip())
            normalized = normalize_transformed_sentence(data)
            return json.dumps(normalized)
        except json.JSONDecodeError as e:
            app_logger.error(
                "[SentenceTransformer] Invalid JSON input provided. "
                f"reason={e} input_text={text}"
            )
            return text

    try:
        gateway = LlmGatewayFactory.create_sentence_decomposer_and_transformer()

        sanitized_text = text.replace('"', "").strip()
        if sanitized_text.startswith(("- ", "* ", "• ")):
            sanitized_text = sanitized_text[2:].strip()
        sanitized_text = re.sub(r"^\d+[\.)]\s+", "", sanitized_text)

        history_pairs = list(previous_messages_stack.get_history_pairs())
        if history_pairs:
            last_role, last_content = history_pairs[-1]
            if last_role == "user" and last_content and last_content.strip() == sanitized_text:
                history_pairs = history_pairs[:-1]
        sanitized_text = _resolve_pronouns_with_history(sanitized_text, history_pairs)
        history = []
        for role, content in history_pairs:
            if not content or not content.strip():
                continue
            label = "Q" if role == "user" else "A"
            history.append(f"{label}: {content.strip()}")
        instruction = _load_sentence_transformer_prompt()
        try:
            gateway.set_system_prompt(instruction)
        except AttributeError:
            pass

        if history:
            history_text = "\n".join(history)
            user_message = (
                f"Conversation history (for context only): {history_text}\n\n"
                f"Sentence: {sanitized_text}"
            )
        else:
            user_message = f"Sentence: {sanitized_text}"

        app_logger.info(
            "[SentenceTransformer] LLM call\n"
            f"Input:\n{user_message}\n"
        )
        transformed = gateway.generate_completion(user_message)
        app_logger.info(
            "[SentenceTransformer] LLM output\n"
            f"Output:\n{parser.format_for_log(transformed)}\n"
        )
        try:
            data = parser.parse(transformed)
        except json.JSONDecodeError as e:
            app_logger.warning(
                "[SentenceTransformer] Invalid JSON from LLM, retrying once. "
                f"reason={e} input_text={text}"
            )
            retry_transformed = gateway.generate_completion(user_message)
            app_logger.info(
                "[SentenceTransformer] LLM retry output\n"
                f"Output:\n{parser.format_for_log(retry_transformed)}\n"
            )
            try:
                data = parser.parse(retry_transformed)
                transformed = retry_transformed
            except json.JSONDecodeError as retry_error:
                app_logger.error(
                    "[SentenceTransformer] LLM returned invalid JSON. "
                    f"reason={retry_error} input_text={text} history={history} "
                    f"user_message={user_message} llm_output={retry_transformed}"
                )
                return text
        normalized = normalize_transformed_sentence(data)

        return json.dumps(normalized)
    except Exception as e:
        app_logger.error(
            "[SentenceTransformer] Error transforming sentence with LLM. "
            f"reason={e} input_text={text}"
        )
        return text


def _resolve_pronouns_with_history(sentence: str, history_pairs: list[tuple[str, str]]) -> str:
    if not sentence or not history_pairs:
        return sentence
    last_subject = _extract_last_subject(history_pairs)
    if not last_subject:
        return sentence
    replacements = {
        " he ": f" {last_subject} ",
        " she ": f" {last_subject} ",
        " they ": f" {last_subject} ",
        " him ": f" {last_subject} ",
        " her ": f" {last_subject} ",
    }
    padded = f" {sentence.strip()} "
    lowered = padded.lower()
    for token, replacement in replacements.items():
        if token in lowered:
            lowered = lowered.replace(token, replacement)
    return lowered.strip()


def _extract_last_subject(history_pairs: list[tuple[str, str]]) -> str | None:
    for role, content in reversed(history_pairs):
        if role != "user" or not content:
            continue
        text = content.strip()
        for sep in (" is ", " are "):
            if sep in text:
                return text.split(sep)[0].strip()
        return text.split()[0] if text else None
    return None



from __future__ import annotations

from typing import List, Optional, Dict, Tuple, Iterable
import math
import operator
import re

from analogai.foundation.settings import config as infra_config
from analogai.deepthink.messaging.fact.fact import Fact
from analogai.deepthink.messaging.fact.enums.fact_reference_type import FactReferenceType
from analogai.deepthink.messaging.fact.repositories.fact_repository import FactRepositoryPort
from analogai.deepthink.messaging.fact.repositories.infrastructure.fact_storage_mapper import FactStorageMapper


class _RepositoryFactResolver:
    def __init__(self, repository: "FactRepositoryPort") -> None:
        self._repository = repository

    def resolve(self, fact_id: str) -> Optional[Fact]:
        return self._repository.get_fact(fact_id)


class InMemoryFactRepository(FactRepositoryPort):
    def __init__(self) -> None:
        self._facts: Dict[str, Fact] = {}
        self._facts_by_turn: Dict[str, List[str]] = {}
        self._turn_index: Dict[Tuple[str, str], List[str]] = {}
        self._statement_links: Dict[str, str] = {}
        self._belief_links: Dict[str, str] = {}
        self._notion_links: Dict[str, str] = {}
        self._hypothetical_links: Dict[str, str] = {}
        self._resolver = _RepositoryFactResolver(self)

    def register_turn(self, user_id: str, agent_id: str, turn_id: str) -> None:
        self._turn_index.setdefault((user_id, agent_id), []).append(turn_id)

    def save_fact(self, fact: Fact) -> None:
        self._facts[fact.id] = fact
        if fact.turn_id:
            self._facts_by_turn.setdefault(fact.turn_id, []).append(fact.id)

    def get_fact(self, id: str) -> Optional[Fact]:
        fact = self._facts.get(id)
        return self._attach_resolver(fact) if fact else None

    def update_facts_statement_id(self, turn_id: str, statement_id: str) -> None:
        ids = self._facts_by_turn.get(turn_id, [])
        for fact_key in ids:
            fact = self._facts.get(fact_key)
            if fact is None:
                continue
            self._facts[fact_key] = Fact(
                id=fact.id,
                turn_id=fact.turn_id,
                statement_id=statement_id,
                notion_id=fact.notion_id,
                reference_type=FactReferenceType.STATEMENT.value,
                user_id=fact.user_id,
                agent_id=fact.agent_id,
                text=fact.text,
                embedding=fact.embedding,
                prev_id=fact.prev_id,
                next_id=fact.next_id,
            )

    def update_facts_notion_id(self, turn_id: str, notion_id: str) -> None:
        ids = self._facts_by_turn.get(turn_id, [])
        for fact_key in ids:
            fact = self._facts.get(fact_key)
            if fact is None:
                continue
            self._facts[fact_key] = Fact(
                id=fact.id,
                turn_id=fact.turn_id,
                statement_id=fact.statement_id,
                notion_id=notion_id,
                reference_type=FactReferenceType.NOTION.value,
                user_id=fact.user_id,
                agent_id=fact.agent_id,
                text=fact.text,
                embedding=fact.embedding,
                prev_id=fact.prev_id,
                next_id=fact.next_id,
            )

    def update_facts_hypothetical_statement_id(self, turn_id: str, statement_id: str) -> None:
        ids = self._facts_by_turn.get(turn_id, [])
        for fact_key in ids:
            fact = self._facts.get(fact_key)
            if fact is None:
                continue
            self._facts[fact_key] = Fact(
                id=fact.id,
                turn_id=fact.turn_id,
                statement_id=statement_id,
                notion_id=fact.notion_id,
                reference_type=FactReferenceType.HYPOTHETICAL_STATEMENT.value,
                user_id=fact.user_id,
                agent_id=fact.agent_id,
                text=fact.text,
                embedding=fact.embedding,
                prev_id=fact.prev_id,
                next_id=fact.next_id,
            )

    def get_facts_for_turn(self, turn_id: str) -> List[Fact]:
        return [
            self._attach_resolver(self._facts[cid])
            for cid in self._facts_by_turn.get(turn_id, [])
            if cid in self._facts
        ]

    def get_turn_ids(self, user_id: str, agent_id: str) -> List[str]:
        return list(self._turn_index.get((user_id, agent_id), []))

    def link_statement(self, statement_id: str, turn_id: str) -> None:
        if statement_id and turn_id:
            self._statement_links[statement_id] = turn_id

    def link_belief(self, belief_id: str, turn_id: str) -> None:
        if belief_id and turn_id:
            self._belief_links[belief_id] = turn_id

    def link_notion(self, notion_id: str, turn_id: str) -> None:
        if notion_id and turn_id:
            self._notion_links[notion_id] = turn_id

    def link_hypothetical_statement(self, statement_id: str, turn_id: str) -> None:
        if statement_id and turn_id:
            self._hypothetical_links[statement_id] = turn_id

    def get_turn_id_for_statement(self, statement_id: str) -> Optional[str]:
        return self._statement_links.get(statement_id)

    def get_turn_id_for_belief(self, belief_id: str) -> Optional[str]:
        return self._belief_links.get(belief_id)

    def get_turn_id_for_notion(self, notion_id: str) -> Optional[str]:
        return self._notion_links.get(notion_id)

    def get_turn_id_for_hypothetical_statement(self, statement_id: str) -> Optional[str]:
        return self._hypothetical_links.get(statement_id)

    def search_text(self, user_id: str, agent_id: str, query: str, top_k: int) -> List[Fact]:
        query_lower = query.lower()
        matches: List[Fact] = []
        for fact in self._iter_user_facts(user_id, agent_id):
            if query_lower in (fact.text or "").lower():
                matches.append(self._attach_resolver(fact))
        return matches[: max(1, int(top_k))]

    def search_similar(
        self,
        user_id: str,
        agent_id: str,
        embedding: List[float],
        top_k: int,
        similarity_cutoff: Optional[float] = None,
    ) -> List[Fact]:
        scored: List[Tuple[float, Fact]] = []
        for fact in self._iter_user_facts(user_id, agent_id):
            if not fact.embedding:
                raise RuntimeError("Fact missing embedding")
            score = _cosine_similarity(embedding, fact.embedding)
            if similarity_cutoff is not None and score < similarity_cutoff:
                continue
            scored.append((score, self._attach_resolver(fact)))
        scored.sort(key=operator.itemgetter(0), reverse=True)
        return [fact for _, fact in scored[: max(1, int(top_k))]]

    def _iter_user_facts(self, user_id: str, agent_id: str) -> Iterable[Fact]:
        for fact in self._facts.values():
            if fact.user_id == user_id and fact.agent_id == agent_id:
                yield fact

    def _attach_resolver(self, fact: Fact) -> Fact:
        fact._resolver = self._resolver
        return fact


class MongoFactRepository(FactRepositoryPort):
    def __init__(
        self,
        uri: Optional[str] = None,
        database: Optional[str] = None,
        facts_collection: str = "facts",
        links_collection: str = "fact_links",
        turn_index_collection: str = "fact_turn_index",
    ) -> None:
        try:
            from pymongo import MongoClient
        except Exception as exc:  # pragma: no cover - optional dependency
            raise ImportError("pymongo is required for MongoFactRepository") from exc

        mongo_uri = uri or infra_config.get("MONGODB_URI", "")
        mongo_db = database or infra_config.get("MONGODB_DATABASE", "")
        if not mongo_uri or not mongo_db:
            raise ValueError("MongoDB configuration missing (MONGODB_URI/MONGODB_DATABASE)")

        client = MongoClient(mongo_uri)
        db = client[mongo_db]
        self._facts = db[facts_collection]
        self._links = db[links_collection]
        self._turn_index = db[turn_index_collection]
        self._resolver = _RepositoryFactResolver(self)

    def register_turn(self, user_id: str, agent_id: str, turn_id: str) -> None:
        key = f"{user_id}:{agent_id}"
        self._turn_index.update_one(
            {"_id": key},
            {"$push": {"turn_ids": turn_id}},
            upsert=True,
        )

    def save_fact(self, fact: Fact) -> None:
        doc = FactStorageMapper.to_document(fact)
        self._facts.update_one({"_id": fact.id}, {"$set": doc}, upsert=True)

    def get_fact(self, id: str) -> Optional[Fact]:
        doc = self._facts.find_one({"_id": id})
        fact = FactStorageMapper.from_document(doc)
        if fact is None:
            return None
        return self._attach_resolver(fact)

    def update_facts_statement_id(self, turn_id: str, statement_id: str) -> None:
        if not turn_id or not statement_id:
            return
        self._facts.update_many(
            {"turn_id": turn_id},
            {"$set": {"statement_id": statement_id, "reference_type": FactReferenceType.STATEMENT.value}},
        )

    def update_facts_notion_id(self, turn_id: str, notion_id: str) -> None:
        if not turn_id or not notion_id:
            return
        self._facts.update_many(
            {"turn_id": turn_id},
            {"$set": {"notion_id": notion_id, "reference_type": FactReferenceType.NOTION.value}},
        )

    def update_facts_hypothetical_statement_id(self, turn_id: str, statement_id: str) -> None:
        if not turn_id or not statement_id:
            return
        self._facts.update_many(
            {"turn_id": turn_id},
            {"$set": {"statement_id": statement_id, "reference_type": FactReferenceType.HYPOTHETICAL_STATEMENT.value}},
        )

    def get_facts_for_turn(self, turn_id: str) -> List[Fact]:
        cursor = self._facts.find({"turn_id": turn_id})
        results: List[Fact] = []
        for doc in cursor:
            fact = FactStorageMapper.from_document(doc)
            if fact is None:
                continue
            results.append(self._attach_resolver(fact))
        return results

    def get_turn_ids(self, user_id: str, agent_id: str) -> List[str]:
        key = f"{user_id}:{agent_id}"
        doc = self._turn_index.find_one({"_id": key})
        if not doc:
            return []
        return [str(item) for item in doc.get("turn_ids", []) if item]

    def link_statement(self, statement_id: str, turn_id: str) -> None:
        if statement_id and turn_id:
            self._set_link("statement", statement_id, turn_id)

    def link_belief(self, belief_id: str, turn_id: str) -> None:
        if belief_id and turn_id:
            self._set_link("belief", belief_id, turn_id)

    def link_notion(self, notion_id: str, turn_id: str) -> None:
        if notion_id and turn_id:
            self._set_link("notion", notion_id, turn_id)

    def link_hypothetical_statement(self, statement_id: str, turn_id: str) -> None:
        if statement_id and turn_id:
            self._set_link("hypothetical_statement", statement_id, turn_id)

    def get_turn_id_for_statement(self, statement_id: str) -> Optional[str]:
        return self._get_link("statement", statement_id)

    def get_turn_id_for_belief(self, belief_id: str) -> Optional[str]:
        return self._get_link("belief", belief_id)

    def get_turn_id_for_notion(self, notion_id: str) -> Optional[str]:
        return self._get_link("notion", notion_id)

    def get_turn_id_for_hypothetical_statement(self, statement_id: str) -> Optional[str]:
        return self._get_link("hypothetical_statement", statement_id)

    def search_text(self, user_id: str, agent_id: str, query: str, top_k: int) -> List[Fact]:
        safe = re.escape(query)
        cursor = self._facts.find(
            {
                "user_id": user_id,
                "agent_id": agent_id,
                "text": {"$regex": safe, "$options": "i"},
            }
        ).limit(max(1, int(top_k)))
        results: List[Fact] = []
        for doc in cursor:
            fact = FactStorageMapper.from_document(doc)
            if fact is None:
                continue
            results.append(self._attach_resolver(fact))
        return results

    def search_similar(
        self,
        user_id: str,
        agent_id: str,
        embedding: List[float],
        top_k: int,
        similarity_cutoff: Optional[float] = None,
    ) -> List[Fact]:
        cursor = self._facts.find(
            {
                "user_id": user_id,
                "agent_id": agent_id,
                "embedding": {"$ne": None},
            }
        )
        scored: List[Tuple[float, Fact]] = []
        for doc in cursor:
            fact = FactStorageMapper.from_document(doc)
            if fact is None:
                continue
            fact = self._attach_resolver(fact)
            if not fact.embedding:
                raise RuntimeError("Fact missing embedding")
            score = _cosine_similarity(embedding, fact.embedding)
            if similarity_cutoff is not None and score < similarity_cutoff:
                continue
            scored.append((score, fact))
        scored.sort(key=operator.itemgetter(0), reverse=True)
        return [fact for _, fact in scored[: max(1, int(top_k))]]

    def _set_link(self, kind: str, entity_id: str, turn_id: str) -> None:
        key = f"{kind}:{entity_id}"
        self._links.update_one(
            {"_id": key},
            {"$set": {"turn_id": turn_id, "kind": kind, "entity_id": entity_id}},
            upsert=True,
        )

    def _get_link(self, kind: str, entity_id: str) -> Optional[str]:
        key = f"{kind}:{entity_id}"
        doc = self._links.find_one({"_id": key})
        if not doc:
            return None
        return doc.get("turn_id")

    def _attach_resolver(self, fact: Fact) -> Fact:
        fact._resolver = self._resolver
        return fact


def _cosine_similarity(a: List[float], b: List[float]) -> float:
    if not a or not b:
        return 0.0
    if len(a) != len(b):
        return 0.0
    dot = sum(x * y for x, y in zip(a, b))
    norm_a = math.sqrt(sum(x * x for x in a))
    norm_b = math.sqrt(sum(y * y for y in b))
    if norm_a == 0.0 or norm_b == 0.0:
        return 0.0
    return dot / (norm_a * norm_b)



from typing import Iterator, Optional, Tuple

from analogai.foundation.common.enums import RelationshipType


def _normalize_tag_key(tag_text: str) -> str:
    return tag_text.strip().lower().replace(" ", "_")


def tag_key_to_relationship_type(tag_key: str) -> Optional[RelationshipType]:
    normalized_key = _normalize_tag_key(tag_key)
    if normalized_key.startswith("is_") and normalized_key not in ("is", "is_a", "isa"):
        normalized_key = normalized_key[3:]
    for relationship_type in RelationshipType:
        if relationship_type.name.lower() == normalized_key:
            return relationship_type
        if _normalize_tag_key(relationship_type.value) == normalized_key:
            return relationship_type
    return None


def relationship_type_tag_keys() -> Iterator[Tuple[str, RelationshipType]]:
    for relationship_type in RelationshipType:
        yield relationship_type.value, relationship_type
        key_from_name = relationship_type.name.lower()
        if key_from_name != relationship_type.value:
            yield key_from_name, relationship_type



from typing import cast, TYPE_CHECKING
from analogai.deepthink.presentation.interpreters.interpreter_result import InterpreterResult
from analogai.deepthink.presentation.interpreters.interpreter_registry import InterpreterRegistry
from analogai.deepthink.presentation.interpreters.result_resolver import InterpreterResultResolver

if TYPE_CHECKING:
    from analogai.deepthink.presentation.factories.interpreters_factory import InterpretersFactory


class InterpreterPipeline:
    def __init__(self, interpreters_factory: "InterpretersFactory"):
        self._interpreters_factory = interpreters_factory
        self._interpreters: InterpreterRegistry | None = None
        self._result_resolver = InterpreterResultResolver()

    def _get_interpreters(self) -> InterpreterRegistry:
        if self._interpreters is None:
            try:
                self._interpreters = self._interpreters_factory.create_registry()
            except AttributeError:
                self._interpreters = InterpreterRegistry(self._interpreters_factory.create_all())
        return self._interpreters

    def run(self, parsed_message, user_agent, message_text: str) -> InterpreterResult | None | bool:
        for interpreter in self._get_interpreters().all():
            result = interpreter.interpret(parsed_message, user_agent, message_text)
            resolved = self._result_resolver.resolve(result)
            if self._result_resolver.is_resolved(resolved):
                return cast(InterpreterResult | bool | None, resolved)
        return None




from typing import List

from analogai.deepthink.presentation.parsers.parsed_sentence import ParsedSentence
from analogai.deepthink.presentation.parsers.pipeline.message_parsing_pipeline import (
    MessageParsingPipeline,
)
from analogai.deepthink.presentation.parsers.utils.json_format_validator import JsonFormatValidator


def parse_json(json_str: str) -> List[ParsedSentence]:
    return MessageParsingPipeline().parse(json_str)


def is_json_format(text: str) -> bool:
    return JsonFormatValidator().is_json_like(text)



from __future__ import annotations

from abc import ABC, abstractmethod
from datetime import datetime
from typing import Optional, List
from uuid import uuid4

from analogai.foundation.infrastructure.services.embedding_service import EmbeddingService
from analogai.deepthink.messaging.fact.fact import Fact
from analogai.deepthink.messaging.fact.repositories.fact_repository import FactRepositoryPort
from analogai.deepthink.messaging.fact.repositories.infrastructure.fact_repository import (
    InMemoryFactRepository,
    MongoFactRepository,
)
from analogai.deepthink.messaging.fact.enums.fact_search_mode import FactSearchMode
from analogai.deepthink.messaging.fact.enums.fact_embedding_mode import FactEmbeddingMode
from analogai.deepthink.messaging.text.sentence_chunker import split_into_sentences


class FactServicePort(ABC):
    @abstractmethod
    def record_turn(
        self,
        user_id: str,
        agent_id: str,
        text: str,
        role: str,
        user_name: Optional[str] = None,
        timestamp: Optional[datetime] = None,
    ) -> str:
        raise NotImplementedError

    @abstractmethod
    def link_statement(self, statement_id: str, turn_id: str) -> None:
        raise NotImplementedError

    @abstractmethod
    def link_belief(self, belief_id: str, turn_id: str) -> None:
        raise NotImplementedError

    @abstractmethod
    def link_notion(self, notion_id: str, turn_id: str) -> None:
        raise NotImplementedError

    @abstractmethod
    def link_hypothetical_statement(self, statement_id: str, turn_id: str) -> None:
        raise NotImplementedError

    @abstractmethod
    def get_facts_for_turn(self, turn_id: str) -> List[Fact]:
        raise NotImplementedError

    @abstractmethod
    def get_facts_for_statement(self, statement_id: str) -> List[Fact]:
        raise NotImplementedError

    @abstractmethod
    def get_facts_for_belief(self, belief_id: str) -> List[Fact]:
        raise NotImplementedError

    @abstractmethod
    def get_facts_for_notion(self, notion_id: str) -> List[Fact]:
        raise NotImplementedError

    @abstractmethod
    def get_facts_for_hypothetical_statement(self, statement_id: str) -> List[Fact]:
        raise NotImplementedError

    @abstractmethod
    def get_nearby_turn_facts(
        self,
        turn_id: str,
        user_id: str,
        agent_id: str,
        window: int = 2,
    ) -> List[Fact]:
        raise NotImplementedError

    @abstractmethod
    def search_facts(
        self,
        user_id: str,
        agent_id: str,
        query: str,
        top_k: int = 5,
        similarity_cutoff: Optional[float] = None,
        mode: FactSearchMode = FactSearchMode.VECTOR_THEN_TEXT,
    ) -> List[Fact]:
        raise NotImplementedError


class RepositoryBackedFactService(FactServicePort):
    def __init__(
        self,
        repository: FactRepositoryPort,
        embedding_service: Optional[EmbeddingService] = None,
        embedding_mode: FactEmbeddingMode = FactEmbeddingMode.ENABLED,
        max_chunk_chars: int = 800,
        overlap_sentences: int = 1,
    ) -> None:
        self._repository = repository
        self._embedding_service = embedding_service
        self._embedding_mode = embedding_mode
        self._max_chunk_chars = max(200, int(max_chunk_chars))
        self._overlap_sentences = max(0, int(overlap_sentences))

    def record_turn(
        self,
        user_id: str,
        agent_id: str,
        text: str,
        role: str,
        user_name: Optional[str] = None,
        timestamp: Optional[datetime] = None,
    ) -> str:
        turn_id = str(uuid4())
        self._repository.register_turn(user_id, agent_id, turn_id)
        fact_chunks = self._split_text_into_facts(text)
        embeddings = self._embed_chunks(fact_chunks)
        fact_ids = [str(uuid4()) for _ in fact_chunks]
        for index, fact_text in enumerate(fact_chunks):
            embedding_vector = embeddings[index] if index < len(embeddings) else None
            prev_fact_id = fact_ids[index - 1] if index > 0 else None
            next_fact_id = fact_ids[index + 1] if index + 1 < len(fact_ids) else None
            fact = Fact(
                id=fact_ids[index],
                turn_id=turn_id,
                statement_id=None,
                user_id=user_id,
                agent_id=agent_id,
                text=fact_text,
                embedding=embedding_vector,
                prev_id=prev_fact_id,
                next_id=next_fact_id,
            )
            self._repository.save_fact(fact)
        return turn_id

    def link_statement(self, statement_id: str, turn_id: str) -> None:
        self._repository.link_statement(statement_id, turn_id)
        self._repository.update_facts_statement_id(turn_id, statement_id)

    def link_belief(self, belief_id: str, turn_id: str) -> None:
        self._repository.link_belief(belief_id, turn_id)

    def link_notion(self, notion_id: str, turn_id: str) -> None:
        self._repository.link_notion(notion_id, turn_id)
        self._repository.update_facts_notion_id(turn_id, notion_id)

    def link_hypothetical_statement(self, statement_id: str, turn_id: str) -> None:
        self._repository.link_hypothetical_statement(statement_id, turn_id)
        self._repository.update_facts_hypothetical_statement_id(turn_id, statement_id)

    def get_facts_for_turn(self, turn_id: str) -> List[Fact]:
        return self._repository.get_facts_for_turn(turn_id)

    def get_facts_for_statement(self, statement_id: str) -> List[Fact]:
        turn_id = self._repository.get_turn_id_for_statement(statement_id)
        return self._repository.get_facts_for_turn(turn_id) if turn_id else []

    def get_facts_for_belief(self, belief_id: str) -> List[Fact]:
        turn_id = self._repository.get_turn_id_for_belief(belief_id)
        return self._repository.get_facts_for_turn(turn_id) if turn_id else []

    def get_facts_for_notion(self, notion_id: str) -> List[Fact]:
        turn_id = self._repository.get_turn_id_for_notion(notion_id)
        return self._repository.get_facts_for_turn(turn_id) if turn_id else []

    def get_facts_for_hypothetical_statement(self, statement_id: str) -> List[Fact]:
        turn_id = self._repository.get_turn_id_for_hypothetical_statement(statement_id)
        return self._repository.get_facts_for_turn(turn_id) if turn_id else []

    def get_nearby_turn_facts(
        self,
        turn_id: str,
        user_id: str,
        agent_id: str,
        window: int = 2,
    ) -> List[Fact]:
        index = self._repository.get_turn_ids(user_id, agent_id)
        if not index or turn_id not in index:
            return []
        position = index.index(turn_id)
        start = max(0, position - max(0, int(window)))
        end = min(len(index), position + max(0, int(window)) + 1)
        nearby_ids = index[start:end]
        results: List[Fact] = []
        for tid in nearby_ids:
            results.extend(self._repository.get_facts_for_turn(tid))
        return results

    def search_facts(
        self,
        user_id: str,
        agent_id: str,
        query: str,
        top_k: int = 5,
        similarity_cutoff: Optional[float] = None,
        mode: FactSearchMode = FactSearchMode.VECTOR_THEN_TEXT,
    ) -> List[Fact]:
        if not query or not query.strip():
            return []
        if mode == FactSearchMode.TEXT_ONLY:
            return self._repository.search_text(user_id, agent_id, query, top_k=top_k)
        if mode == FactSearchMode.VECTOR_ONLY:
            return self._search_vector_only(user_id, agent_id, query, top_k, similarity_cutoff)
        if mode == FactSearchMode.VECTOR_THEN_TEXT:
            results = self._search_vector_only(user_id, agent_id, query, top_k, similarity_cutoff)
            return results or self._repository.search_text(user_id, agent_id, query, top_k=top_k)
        return []

    def _embed_chunks(self, chunks: List[str]) -> List[List[float]]:
        if not chunks:
            return []
        if self._embedding_mode != FactEmbeddingMode.ENABLED:
            return []
        if self._embedding_service is None:
            raise RuntimeError("Fact embedding service unavailable")
        return self._embedding_service.encode_batch(chunks)

    def _search_vector_only(
        self,
        user_id: str,
        agent_id: str,
        query: str,
        top_k: int,
        similarity_cutoff: Optional[float],
    ) -> List[Fact]:
        if self._embedding_mode != FactEmbeddingMode.ENABLED:
            return []
        if self._embedding_service is None:
            raise RuntimeError("Fact embedding service unavailable")
        query_embedding = self._embedding_service.encode(query)
        if not query_embedding:
            return []
        return self._repository.search_similar(
            user_id,
            agent_id,
            query_embedding,
            top_k=top_k,
            similarity_cutoff=similarity_cutoff,
        )

    def _split_text_into_facts(self, text: str) -> List[str]:
        if not text or not text.strip():
            return []
        sentence_list = split_into_sentences(text)
        if not sentence_list:
            return [text.strip()]
        fact_chunks: List[str] = []
        current_sentences: List[str] = []
        current_length = 0
        for sentence in sentence_list:
            cleaned_sentence = sentence.strip()
            if not cleaned_sentence:
                continue
            extra_length = len(cleaned_sentence) + (1 if current_sentences else 0)
            if current_sentences and current_length + extra_length > self._max_chunk_chars:
                fact_chunks.append(" ".join(current_sentences).strip())
                overlap_sentences = (
                    current_sentences[-self._overlap_sentences :]
                    if self._overlap_sentences
                    else []
                )
                current_sentences = list(overlap_sentences)
                current_length = sum(len(s) for s in current_sentences) + max(0, len(current_sentences) - 1)
            current_sentences.append(cleaned_sentence)
            current_length = sum(len(s) for s in current_sentences) + max(0, len(current_sentences) - 1)
        if current_sentences:
            fact_chunks.append(" ".join(current_sentences).strip())
        return fact_chunks


class InMemoryFactService(RepositoryBackedFactService):
    def __init__(
        self,
        embedding_service: Optional[EmbeddingService] = None,
        embedding_mode: FactEmbeddingMode = FactEmbeddingMode.ENABLED,
    ):
        super().__init__(
            InMemoryFactRepository(),
            embedding_service=embedding_service,
            embedding_mode=embedding_mode,
        )


class MongoFactService(RepositoryBackedFactService):
    def __init__(
        self,
        repository: Optional[FactRepositoryPort] = None,
        embedding_service: Optional[EmbeddingService] = None,
        embedding_mode: FactEmbeddingMode = FactEmbeddingMode.ENABLED,
    ):
        super().__init__(
            repository or MongoFactRepository(),
            embedding_service=embedding_service,
            embedding_mode=embedding_mode,
        )





from __future__ import annotations

from analogai.deepthink.messaging.commands import RememberCommand, SearchCommand
from analogai.deepthink.messaging.flows.chat_flow import ChatFlow
from analogai.deepthink.messaging.question_answering.question_routing_service import (
    QuestionRoutingService,
)


class ChatFlowBuilder:
    def __init__(
        self,
        input_router: QuestionRoutingService,
        remember_command: RememberCommand,
        search_command: SearchCommand,
    ):
        self._input_router = input_router
        self._remember_command = remember_command
        self._search_command = search_command

    def create(self) -> ChatFlow:
        return ChatFlow(
            input_router=self._input_router,
            remember_command=self._remember_command,
            search_command=self._search_command,
        )



from typing import Optional

from pydantic import BaseModel, ConfigDict
from analogai.foundation.common.dtos.notion_expression import NotionExpression
from analogai.deepthink.application.value_objects.user_agent import UserAgent


class MakeNotionStatementRequest(BaseModel):
    model_config = ConfigDict(strict=True, extra="forbid")

    user_agent: UserAgent
    notion_expression: NotionExpression


class MakeNotionStatementResponse(BaseModel):
    model_config = ConfigDict(strict=True, extra="forbid")

    user_agent: UserAgent
    notion_expression: NotionExpression
    notion_id: Optional[str] = None
    saved: bool = True


MakeNotionStatementRequest.model_rebuild()
MakeNotionStatementResponse.model_rebuild()



import os
import sys
from pathlib import Path
from abc import ABC, abstractmethod


from analogai.foundation.infrastructure.storage.neo4j.neo4j_gateway import Neo4jGateway
from analogai.foundation.infrastructure.storage.in_memory.in_memory_storage_adapter_impl import InMemoryStorageAdapterImpl
from analogai.foundation.infrastructure.storage.mongodb.mongodb_storage_adapter_impl import MongoDBStorageAdapterImpl
from analogai.deepthink.settings import config as app_config

class StorageCleaner(ABC):
    @abstractmethod
    def cleanup(self):
        pass

class ProductionEnvironmentStorageCleaner(StorageCleaner):
    def cleanup(self):
        neo4j_gateway = Neo4jGateway()
        with neo4j_gateway.session() as session:
            session.run("MATCH (n) DETACH DELETE n")
        
        storage_adapter = MongoDBStorageAdapterImpl()
        storage_adapter.clear()

class DevEnvironmentStorageCleaner(StorageCleaner):
    def cleanup(self):
        storage_adapter = InMemoryStorageAdapterImpl()
        storage_adapter.clear()

def cleanup_storages():
    environment = app_config.get("ENVIRONMENT")
    cleaner = (
        DevEnvironmentStorageCleaner()
        if environment == "development"
        else ProductionEnvironmentStorageCleaner()
    )
    cleaner.cleanup()

if __name__=="__main__":
    cleanup_storages()





from __future__ import annotations

from typing import Iterable, List

from analogai.deepthink.application.usecases.learning.make_conditional_statement.models import CriterionValue


class AttributeValueNormalizer:
    def normalize(self, attributes: Iterable) -> List[CriterionValue]:
        if not attributes:
            return []
        normalized: List[CriterionValue] = []
        for attr in attributes:
            tag = attr.tag
            values = attr.values or []
            value = values[0] if values else None
            unit = attr.unit
            min_value = attr.min_value
            max_value = attr.max_value
            if tag:
                normalized.append(
                    CriterionValue(
                        tag=tag,
                        value=value,
                        unit=unit,
                        min_value=min_value,
                        max_value=max_value,
                    )
                )
        return normalized
